# cython: language_level=3

from libc.stdint cimport uint32_t, uint64_t
from . cimport cpp

cdef class ClusterEditingSolver:
	cdef cpp.ClusterEditingSolver *thisptr
	cdef TriangleSparseMatrix m


cdef class TriangleSparseMatrix:
	cdef cpp.TriangleSparseMatrix *thisptr


cdef class ReadScoring:
	cdef cpp.ReadScoring *thisptr


cdef class HaploThreader:
	cdef cpp.HaploThreader *thisptr


cdef class SwitchFlipCalculator:
	cdef cpp.SwitchFlipCalculator *thisptr
	cdef uint32_t ploidy



from whatshap.core import PedigreeDPTable, Pedigree, NumericSampleIds, PhredGenotypeLikelihoods
from whatshap.testhelpers import string_to_readset, matrix_to_readset, canonic_index_to_biallelic_gt
from whatshap.verification import verify_mec_score_and_partitioning


def verify(rs, all_heterozygous=False):
    positions = rs.get_positions()
    # recombination costs 1, should not occur
    recombcost = [1] * len(positions)
    pedigree = Pedigree(NumericSampleIds())
    genotype_likelihoods = [
        None if all_heterozygous else PhredGenotypeLikelihoods([0, 0, 0])
    ] * len(positions)
    # all genotypes heterozygous
    pedigree.add_individual(
        "individual0",
        [canonic_index_to_biallelic_gt(1) for _ in range(len(positions))],
        genotype_likelihoods,
    )
    dp_table = PedigreeDPTable(rs, recombcost, pedigree, distrust_genotypes=not all_heterozygous)
    verify_mec_score_and_partitioning(dp_table, rs)


def test_string():
    reads = """
      0             0
      110111111111
      00100
           0001000000
           000
            10100
                  101
    """
    rs = string_to_readset(reads)
    verify(rs, True)
    verify(rs, False)


def test_matrix():
    with open("tests/test.matrix") as f:
        rs = matrix_to_readset(f)
    verify(rs, True)
    verify(rs, False)



# cython: language_level=3

"""
Wrappers for core C++ classes.
"""
# Do not use the distutils directives here, but configure everything in
# setup.py, such as language and sources. It would work during development, but
# during a regular installation, the module will be compiled from the
# pre-generated .cpp file and the .pyx file is not read.

from libcpp cimport bool
from libcpp.string cimport string
from libcpp.vector cimport vector
from libc.stdint cimport uint32_t, uint64_t
from . cimport cpp

from .variant import Variant
from collections import namedtuple
from cython.operator cimport dereference as deref


cdef class NumericSampleIds:
	"""
	Mapping of sample names (strings) to numeric ids.
	"""
	def __cinit__(self):
		self.mapping = {}
		self.frozen = False

	def __getitem__(self, sample):
		if not self.frozen and sample not in self.mapping:
			self.mapping[sample] = len(self.mapping)
		return self.mapping[sample]

	def __len__(self):
		return len(self.mapping)

	def __str__(self):
		return str(self.mapping)

	def freeze(self):
		"""No longer allow modifications"""
		# TODO We should try to have a second class (FrozenNumericSampleIds) for
		# this or try to get rid of NumericSampleIds
		self.frozen = True

	def inverse_mapping(self):
		"""Returns a dict mapping numeric ids to sample names."""
		return {numeric_id:name for name, numeric_id in self.mapping.items()}
	
	def __getstate__(self):
		return (self.mapping, self.frozen)
	
	def __setstate__(self, state):
		mapping, frozen = state
		self.mapping = mapping
		self.frozen = frozen


cdef class Read:
	def __cinit__(self, str name = None, int mapq = 0, int source_id = 0, int sample_id = 0, int reference_start = -1, str BX_tag = None):
		cdef string _name = b''
		cdef string _BX_tag = b''
		if name is None:
			self.thisptr = NULL
			self.ownsptr = False
		else:
			# TODO: Is this the best way to handle string arguments?
			_name = name.encode('UTF-8')
			if BX_tag is not '' and BX_tag is not None:
				_BX_tag = BX_tag.encode('UTF-8')
			self.thisptr = new cpp.Read(_name, mapq, source_id, sample_id, reference_start, _BX_tag)
			self.ownsptr = True

	def __dealloc__(self):
		if self.ownsptr:
			assert self.thisptr != NULL
			del self.thisptr

	def __repr__(self):
		assert self.thisptr != NULL
		return 'Read(name={!r}, mapq={}, source_id={}, sample_id={}, reference_start={},  BX_tag={}, variants={})'.format(
			self.name, self.mapqs, self.source_id, self.sample_id, self.reference_start, self.BX_tag, list(self))

	property mapqs:
		def __get__(self):
			assert self.thisptr != NULL
			return tuple(self.thisptr.getMapqs())

	property name:
		def __get__(self):
			assert self.thisptr != NULL
			return self.thisptr.getName().decode('utf-8')

	property source_id:
		def __get__(self):
			assert self.thisptr != NULL
			return self.thisptr.getSourceID()

	property sample_id:
		def __get__(self):
			assert self.thisptr != NULL
			return self.thisptr.getSampleID()
	
	property reference_start:
		def __get__(self):
			assert self.thisptr != NULL
			return self.thisptr.getReferenceStart()

	property BX_tag:
		def __get__(self):
			assert self.thisptr != NULL
			return self.thisptr.getBXTag().decode('utf-8')

	def __iter__(self):
		"""Iterate over all variants in this read"""
		assert self.thisptr != NULL
		for i in range(len(self)):
			yield self[i]

	def __len__(self):
		"""Return number of variants in this read"""
		assert self.thisptr != NULL
		return self.thisptr.getVariantCount()

	def __getitem__(self, key):
		"""Return Variant object at the given integer index"""
		assert self.thisptr != NULL
		if isinstance(key, slice):
			raise NotImplementedError("Read does not support slices")
		assert isinstance(key, int)
		cdef int n = self.thisptr.getVariantCount()
		if not (-n <= key < n):
			raise IndexError('Index out of bounds: {}'.format(key))
		if key < 0:
			key = n + key
		return Variant(
			position=self.thisptr.getPosition(key),
			allele=self.thisptr.getAllele(key),
			quality=self.thisptr.getVariantQuality(key)
		)

	def __setitem__(self, index, variant):
		assert self.thisptr != NULL
		cdef int n = self.thisptr.getVariantCount()
		if not (-n <= index < n):
			raise IndexError('Index out of bounds: {}'.format(index))
		if index < 0:
			index = n + index
		if not isinstance(variant, Variant):
			raise ValueError('Expected instance of Variant, but found {}'.format(type(variant)))
		self.thisptr.setPosition(index, variant.position)
		self.thisptr.setAllele(index, variant.allele)
		self.thisptr.setVariantQuality(index, variant.quality)

	def __contains__(self, position):
		"""Return whether this read contains a variant at the given position.
		A linear search is used.
		"""
		assert self.thisptr != NULL
		assert isinstance(position, int)
		for variant in self:
			if variant.position == position:
				return True
		return False
	
	def __getstate__(self):
		mapqs = [mapq for mapq in self.mapqs]
		variants = [(var.position, var.allele, var.quality) for var in self]
		return (mapqs, self.name, self.source_id, self.sample_id, self.reference_start, self.BX_tag, variants)
	
	def __setstate__(self, state):
		mapqs, name, source_id, sample_id, reference_start, BX_tag, variants = state
		
		# TODO: Duplicated code from __cinit__ is ugly, but cinit cannot be used here directly
		cdef string _name = b''
		cdef string _BX_tag = b''
		if name is None:
			self.thisptr = NULL
			self.ownsptr = False
		else:
			# TODO: Is this the best way to handle string arguments?
			_name = name.encode('UTF-8')
			if BX_tag is not b'' and BX_tag is not None:
				_BX_tag = BX_tag.encode('UTF-8')
			self.thisptr = new cpp.Read(_name, mapqs[0] if len(mapqs) > 0 else 0, source_id, sample_id, reference_start, _BX_tag)
			self.ownsptr = True

		for mapq in mapqs[1:]:
			self.add_mapq(mapq)
		for (pos, allele, quality) in variants:
			self.add_variant(pos, allele, quality)

	def add_variant(self, int position, int allele, int quality):
		assert self.thisptr != NULL
		self.thisptr.addVariant(position, allele, quality)

	def add_mapq(self, int mapq):
		assert self.thisptr != NULL
		self.thisptr.addMapq(mapq)

	def sort(self):
		assert self.thisptr != NULL
		self.thisptr.sortVariants()

	def is_sorted(self):
		assert self.thisptr != NULL
		return self.thisptr.isSorted()

	def has_BX_tag(self):
		assert self.thisptr != NULL
		return self.thisptr.hasBXTag()


cdef class ReadSet:
	def __cinit__(self):
		self.thisptr = new cpp.ReadSet()

	def __dealloc__(self):
		del self.thisptr

	def add(self, Read read):
		"""Adds a read to the set.
		WARNING: this will internally create a copy of the wrapped C++ Read object,
		so that subsequent changes to the Read don't affect the
		newly created copy that is added to the ReadSet."""
		self.thisptr.add(new cpp.Read(read.thisptr[0]))

	def __str__(self):
		return self.thisptr.toString().decode('utf-8')

	def __iter__(self):
		for i in range(self.thisptr.size()):
			yield self[i]

	def __len__(self):
		return self.thisptr.size()

	def __getitem__(self, key):
		if isinstance(key, slice):
			raise NotImplementedError('ReadSet does not support slices')
		cdef string name = b''
		cdef cpp.Read* cread = NULL
		cdef Read read = Read()
		if isinstance(key, int):
			read.thisptr = self.thisptr.get(key)
		elif isinstance(key, str):
			raise NotImplementedError('Querying a ReadSet by read name is deprecated, please query by (source_id, name) instead')
		elif isinstance(key, tuple) and (len(key) == 2) and (isinstance(key[0],int) and isinstance(key[1],str)):
			source_id = key[0]
			name = key[1].encode('UTF-8')
			cread = self.thisptr.getByName(name, source_id)
			if cread == NULL:
				raise KeyError(key)
			else:
				read.thisptr = cread
		else:
			assert False, 'Invalid key: {}'.format(key)
		return read
	
	def __getstate__(self):
		return ([read for read in self])
	
	def __setstate__(self, state):
		self.thisptr = new cpp.ReadSet()
		for read in state:
			self.add(read)

	#def get_by_name(self, name):
		#cdef string _name = name.encode('UTF-8')
		#cdef cpp.Read* cread = self.thisptr.getByName(_name)
		#cdef Read read = Read()
		#if cread == NULL:
			#raise KeyError(name)
		#else:
			#read.thisptr = cread
			#return read

	def sort(self):
		"""Sort contained reads by the position of the first variant they contain. Note that
		this is not necessarily the variant with the lowest position, unless sort() has been
		called on all contained reads. Ties are resolved by comparing the read name."""
		self.thisptr.sort()

	def subset(self, reads_to_select):
		# TODO: is there a way of avoiding to unecessarily creating/destroying a ReadSet object?
		cdef cpp.IndexSet* index_set = new cpp.IndexSet()
		cdef int i
		for i in reads_to_select:
			index_set.add(i)
		result = ReadSet()
		del result.thisptr
		result.thisptr = self.thisptr.subset(index_set)
		del index_set
		return result

	def get_positions(self):
		cdef vector[unsigned int]* v = self.thisptr.get_positions()
		result = list(v[0])
		del v
		return result


cdef class PedigreeDPTable:
	def __cinit__(self, ReadSet readset, recombcost, Pedigree pedigree, bool distrust_genotypes = False, positions = None):
		"""Build the DP table from the given read set which is assumed to be sorted;
		that is, the variants in each read must be sorted by position and the reads
		in the read set must also be sorted (by position of their left-most variant).
		"""
		cdef vector[unsigned int]* c_positions = NULL
		if positions is not None:
			c_positions = new vector[unsigned int]()
			for pos in positions:
				c_positions.push_back(pos)
		self.thisptr = new cpp.PedigreeDPTable(readset.thisptr, recombcost, pedigree.thisptr, distrust_genotypes, c_positions)
		self.pedigree = pedigree

	def __dealloc__(self):
		del self.thisptr

	def get_super_reads(self):
		"""Obtain optimal-score haplotypes. Returns a triple (mother, father, child)
		IMPORTANT: The ReadSet given at construction time must not have been altered.
		DPTable retained a pointer to this set and will access it again. If it has
		been altered, behavior is undefined.
		TODO: Change that.
		"""
		cdef vector[cpp.ReadSet*]* read_sets = new vector[cpp.ReadSet*]()

		for i in range(len(self.pedigree)):
			read_sets.push_back(new cpp.ReadSet())
		transmission_vector_ptr = new vector[unsigned int]()
		self.thisptr.get_super_reads(read_sets, transmission_vector_ptr)
		
		results = []
		for i in range(read_sets.size()):
			rs = ReadSet()
			del rs.thisptr
			rs.thisptr = deref(read_sets)[i]
			results.append(rs)

		python_transmission_vector = list(transmission_vector_ptr[0])
		del transmission_vector_ptr
		return results, python_transmission_vector

	def get_optimal_cost(self):
		"""Returns the cost resulting from solving the Minimum Error Correction (MEC) problem."""
		return self.thisptr.get_optimal_score()

	def get_optimal_partitioning(self):
		"""Returns a list of the same size as the read set, where each entry is either 0 or 1,
		telling whether the corresponding read is in partition 0 or in partition 1,"""
		cdef vector[bool]* p = self.thisptr.get_optimal_partitioning()
		result = [0 if x else 1 for x in p[0]]
		del p
		return result


cdef class Pedigree:
	def __cinit__(self, numeric_sample_ids):
		self.thisptr = new cpp.Pedigree()
		self.numeric_sample_ids = numeric_sample_ids

	def __dealloc__(self):
		del self.thisptr

	def add_individual(self, id, genotypes, genotype_likelihoods=None):
		cdef vector[cpp.Genotype*] gt_vector
		for gt in genotypes:
			gt_vector.push_back(new cpp.Genotype((<Genotype?>gt).thisptr[0]))
		cdef vector[cpp.PhredGenotypeLikelihoods*] gl_vector
		if genotype_likelihoods:
			for gl in genotype_likelihoods:
				if gl is None:
					gl_vector.push_back(NULL)
				else:
					gl_vector.push_back(new cpp.PhredGenotypeLikelihoods((<PhredGenotypeLikelihoods?>gl).thisptr[0]) )
		else:
			for _ in genotypes:
				gl_vector.push_back(NULL)
		self.thisptr.addIndividual(self.numeric_sample_ids[id], gt_vector, gl_vector)

	def add_relationship(self, father_id, mother_id, child_id):
		self.thisptr.addRelationship(self.numeric_sample_ids[father_id], self.numeric_sample_ids[mother_id], self.numeric_sample_ids[child_id])

	property variant_count:
		"""Number of variants stored for each individual."""
		def __get__(self):
			return self.thisptr.get_variant_count()

	def genotype(self, sample_id, unsigned int variant_index):
		cdef const cpp.Genotype* gt = self.thisptr.get_genotype_by_id(self.numeric_sample_ids[sample_id], variant_index)
		return Genotype(gt[0].as_vector())

	def genotype_likelihoods(self, sample_id, unsigned int variant_index):
		cdef const cpp.PhredGenotypeLikelihoods* gl = self.thisptr.get_genotype_likelihoods_by_id(self.numeric_sample_ids[sample_id], variant_index)
		if gl == NULL:
			return None
		else:
			return PhredGenotypeLikelihoods(gl[0].as_vector(), gl[0].get_ploidy(), gl[0].get_nr_alleles())

	def __len__(self):
		return self.thisptr.size()

	def __str__(self):
		return self.thisptr.toString().decode('utf-8')


cdef class PhredGenotypeLikelihoods:
	def __cinit__(self, vector[double] gl, unsigned int ploidy=2, unsigned int nr_alleles=2):
		self.thisptr = new cpp.PhredGenotypeLikelihoods(gl, ploidy, nr_alleles)

	def __dealloc__(self):
		del self.thisptr

	def __str__(self):
		return self.thisptr.toString().decode('utf-8')

	def __getitem__(self, Genotype genotype):
		assert self.thisptr != NULL
		assert genotype.is_diploid_and_biallelic()
		return self.thisptr.get(genotype.thisptr[0])

	def __len__(self):
		return self.thisptr.size()

	def __iter__(self):
		for genotype in self.genotypes():
			yield self[genotype]
			
	def __eq__(self, PhredGenotypeLikelihoods other):
		if self.genotypes() != other.genotypes():
			return False
		for genotype in self.genotypes():
			if self[genotype] != other[genotype]:
				return False
		return True
		
	def genotypes(self):
		cdef vector[cpp.Genotype]* genotypes = new vector[cpp.Genotype]()
		self.thisptr.get_genotypes(deref(genotypes))
		result = [Genotype(genotype.as_vector()) for genotype in genotypes[0]]
		del genotypes
		return result
	
	
def binomial_coefficient(int n, int k):
	return cpp.binomial_coefficient(n, k)

			
cdef class Genotype:
	
	def __cinit__(self, vector[uint32_t] alleles):
		self.thisptr = new cpp.Genotype(alleles)
		self.ploidy = self.thisptr.get_ploidy()
		self.index = self.thisptr.get_index()

	def __dealloc__(self):
		del self.thisptr

	def __str__(self):
		return self.thisptr.toString().decode('utf-8')
	
	def __repr__(self):
		return self.thisptr.toString().decode('utf-8')

	def is_none(self):
		return self.thisptr.is_none()

	def get_index(self):
		return self.thisptr.get_index()

	def as_vector(self):
		result = []
		cdef vector[uint32_t] alleles = self.thisptr.as_vector()
		for allele in alleles:
			result.append(allele)
		return alleles

	def is_homozygous(self):
		return self.thisptr.is_homozygous()
	
	def is_diploid_and_biallelic(self):
		return self.thisptr.is_diploid_and_biallelic()
	
	def get_ploidy(self):
		return self.thisptr.get_ploidy()

	def __eq__(self, Genotype g):
		return self.thisptr[0] == g.thisptr[0]

	def __ne__(self, Genotype g):
		return self.thisptr[0] != g.thisptr[0]
	
	def __lt__(self, Genotype g):
		return self.thisptr[0] < g.thisptr[0]
	
	def __hash__(self):
		return hash(tuple(self.alleles))
	
	def __reduce__(self):
		# a tuple as specified in the pickle docs - (class_or_constructor, 
		# (tuple, of, args, to, constructor))
		cdef vector[uint32_t] alleles = cpp.convert_index_to_alleles(self.index, self.ploidy)
		return (self.__class__, tuple([alleles]))
	
	
def get_max_genotype_ploidy():
	return cpp.get_max_genotype_ploidy()


def get_max_genotype_alleles():
	return cpp.get_max_genotype_alleles()


cdef class GenotypeDPTable:
	def __cinit__(self, numeric_sample_ids, ReadSet readset, recombcost, Pedigree pedigree, positions = None):
		"""Build the DP table from the given read set which is assumed to be sorted;
		that is, the variants in each read must be sorted by position and the reads
		in the read set must also be sorted (by position of their left-most variant).
		"""
		cdef vector[unsigned int]* c_positions = NULL
		if positions is not None:
			c_positions = new vector[unsigned int]()
			for pos in positions:
				c_positions.push_back(pos)
		self.thisptr = new cpp.GenotypeDPTable(readset.thisptr, recombcost, pedigree.thisptr, c_positions)
		self.pedigree = pedigree
		self.numeric_sample_ids = numeric_sample_ids

	def __dealloc__(self):
		del self.thisptr

	def get_genotype_likelihoods(self, sample_id, unsigned int pos):
		return PhredGenotypeLikelihoods(self.thisptr.get_genotype_likelihoods(self.numeric_sample_ids[sample_id],pos))


def compute_genotypes(ReadSet readset, positions = None):
	cdef vector[cpp.Genotype]* genotypes_vector = new vector[cpp.Genotype]()
	cdef vector[cpp.GenotypeDistribution]* gl_vector = new vector[cpp.GenotypeDistribution]()
	cdef vector[unsigned int]* c_positions = NULL
	if positions is not None:
		c_positions = new vector[unsigned int]()
		for pos in positions:
			c_positions.push_back(pos)
	cpp.compute_genotypes(readset.thisptr[0], genotypes_vector, gl_vector, c_positions)
	# TODO: Inefficient
	genotypes = [Genotype(genotype.as_vector()) for genotype in genotypes_vector[0]]
	#genotypes = list(genotypes_vector[0])
	gls = [(gl_vector[0][i].probabilityOf(0), gl_vector[0][i].probabilityOf(1), gl_vector[0][i].probabilityOf(2)) for i in range(gl_vector[0].size())]
	del genotypes_vector
	del gl_vector
	return genotypes, gls


cdef class HapChatCore:
	def __cinit__(self, ReadSet readset):
		self.thisptr = new cpp.HapChatCore(readset.thisptr)
	def __dealloc__(self):
		del self.thisptr
	def get_length(self):
		return self.thisptr.get_length()
	def get_super_reads(self):
		cdef vector[cpp.ReadSet*]* read_sets = new vector[cpp.ReadSet*]()
		leng=self.thisptr.get_length()
		for i in range(leng):
			read_sets.push_back(new cpp.ReadSet())
		self.thisptr.get_super_reads(read_sets)
		
		results = []
		for i in range(read_sets.size()):
			rs = ReadSet()
			del rs.thisptr
			rs.thisptr = deref(read_sets)[i]
			results.append(rs)
		
		return results, None
	def get_optimal_cost(self):
		return self.thisptr.get_optimal_cost()
	def get_optimal_partitioning(self):
		cdef vector[bool]* p = self.thisptr.get_optimal_partitioning()
		result = ['*' for x in p[0]]
		del p
		return result



import pytest
from whatshap.pedigree import GeneticMapRecombinationCostComputer, ParseError


def test_read_genetic_map(tmp_path):
    path = tmp_path / "genetic.map"
    path.write_text("ignored header\n" "568527 0 0\n" "723891 2.9813105581 0.417644215424158\n")
    _ = GeneticMapRecombinationCostComputer(str(path))


def test_read_wrong_number_of_fields(tmp_path):
    path = tmp_path / "genetic.map"
    path.write_text(
        "ignored header\n" "55550 0 0\n" "568322 0 0 17\n" "723891 2.9813105581 0.417644215424158\n"
    )
    with pytest.raises(ParseError):
        _ = GeneticMapRecombinationCostComputer(str(path))


def test_invalid_int(tmp_path):
    path = tmp_path / "genetic.map"
    path.write_text("ignored header\n" "55550 0 0\n" "abc 0 0\n")
    with pytest.raises(ParseError):
        _ = GeneticMapRecombinationCostComputer(str(path))


def test_invalid_float(tmp_path):
    path = tmp_path / "genetic.map"
    path.write_text("ignored header\n" "55550 0 abc\n")
    with pytest.raises(ParseError):
        _ = GeneticMapRecombinationCostComputer(str(path))



"""
Find connected components.
"""
from abc import abstractmethod
from collections import OrderedDict
from typing import TypeVar, Generic, Optional, Iterable
import typing

if typing.TYPE_CHECKING:
    from typing_extensions import Protocol
else:
    Protocol = object


T = TypeVar("T")
C = TypeVar("C", bound="Comparable")


class Comparable(Protocol):
    @abstractmethod
    def __lt__(self: C, other: C) -> bool:
        pass


class Node(Generic[C]):
    def __init__(self, value: C, parent: Optional["Node"]):
        self.value = value
        self.parent = parent

    def __repr__(self):
        return f"Node(value={self.value}, parent={self.parent})"


class ComponentFinder(Generic[C]):
    """
    Find connected components. A ComponentFinder is initialized with a list of
    values. These are initially partitioned such that each value is in a
    separate set. By calling merge(x, y), the two sets containing values x and
    y are merged. Calling find(x) returns a "representative" value of the set
    that value x is in. x and y are in the same set iff find(x) == find(y).
    The representative is always the minimum value of the set.

    This implements a variant of the Union-Find algorithm, but without the
    "union by rank" strategy since we want the smallest node to be the
    representative. It could perhaps be optimized, but this function is not
    the current bottleneck.
    """

    def __init__(self, values: Iterable[C]):
        self.nodes = {value: Node(value, None) for value in values}

    def merge(self, x: C, y: C) -> None:
        assert x != y
        x_root = self._find_node(x)
        y_root = self._find_node(y)

        if x_root is y_root:
            return

        # Merge while making sure that the node with the smaller value is the
        # new parent.
        if x_root.value < y_root.value:
            y_root.parent = x_root
        else:
            x_root.parent = y_root

    def _find_node(self, value: C) -> Node[C]:
        node = root = self.nodes[value]
        while root.parent is not None:
            root = root.parent

        # compression path
        while node.parent is not None:
            node.parent, node = root, node.parent
        return root

    def find(self, value: C) -> C:
        """
        Return which component x belongs to, identified by the smallest value.
        """
        return self._find_node(value).value

    def print(self):
        for x in sorted(self.nodes):
            print(x, ":", self.nodes[x], "is represented by", self._find_node(x))


class CyclicGraphError(Exception):
    pass


class Graph:
    """Directed graph that can sort topologically"""

    def __init__(self):
        # map node to a list of neighbors
        self._neighbors = OrderedDict()

    def add_edge(self, node1, node2):
        """The edge is directed from node1 to node2"""
        if node1 not in self._neighbors:
            self._neighbors[node1] = []
        self._neighbors[node1].append(node2)
        if node2 not in self._neighbors:
            self._neighbors[node2] = []

    def toposorted(self):
        """
        Return nodes of the graph sorted topologically.
        For all edges u -> v that the graph has, node v will appear
        before node u.
        """
        order = []
        colors = {node: "white" for node in self._neighbors}

        def visit(node):
            assert colors[node] == "white"
            colors[node] = "gray"
            for neighbor in self._neighbors[node]:
                if colors[neighbor] == "white":
                    visit(neighbor)
                elif colors[neighbor] == "gray":
                    raise CyclicGraphError(f"Cycle involving {node!r} and {neighbor!r} detected")
            order.append(node)
            colors[node] = "black"

        for node in self._neighbors:
            if colors[node] == "white":
                visit(node)
        return order



"""
Split reads by haplotype

This subcommand reads either a FASTQ or a BAM file and a list of haplotype assignments
(such as generated by whatshap haplotag --output-haplotag-list); it then outputs one
FASTQ or BAM per haplotype.

1. BAM mode is intended for unmapped BAMs (such as provided by PacBio).
2. The output format is the same as the input format (FASTQ or BAM).
  (Reading BAM but writing FASTQ or vice versa is not possible.)

Examples:

    whatshap split --output-h1 h1.fastq.gz --output-h2 h2.fastq.gz reads.fastq.gz haplotypes.txt
    whatshap split --output-h1 h1.bam --output-h2 h2.bam reads.bam haplotypes.txt
"""
import logging
import os
import pysam
from collections import defaultdict, Counter
import itertools

from xopen import xopen

from contextlib import ExitStack
from whatshap.utils import detect_file_format
from whatshap.timer import StageTimer

logger = logging.getLogger(__name__)


# fmt: off
def add_arguments(parser):
    arg = parser.add_argument
    arg('--output-h1', default=None,
        help='Output file to write reads from Haplotype 1 to. Use ending .gz to '
        'create gzipped file.')
    arg('--output-h2', default=None,
        help='Output file to write reads from Haplotype 2 to. Use ending .gz to '
        'create gzipped file.')
    arg('--output-untagged', default=None,
        help='Output file to write untagged reads to. Use ending .gz to '
        'create gzipped file.')
    arg('--add-untagged', default=False, action='store_true',
        help='Add reads without tag to both H1 and H2 output streams.')
    arg('--only-largest-block', default=False, action='store_true',
        help='Only consider reads to be tagged if they belong to the largest '
        'phased block (in terms of read count) on their respective chromosome')
    arg('--discard-unknown-reads', default=False, action='store_true',
        help='Only check the haplotype of reads listed in the haplotag list file. '
        'Reads (read names) not contained in this file will be discarded. '
        'In the default case (= keep unknown reads), those reads would be '
        'considered untagged and end up in the respective output file. '
        'Please be sure that the read names match between the input FASTQ/BAM '
        'and the haplotag list file.')
    arg('--read-lengths-histogram', default=None,
        help='Output file to write read lengths histogram to in tab separated format.')
    arg('reads_file', metavar='READS', help='Input FASTQ/BAM file with reads (FASTQ can be gzipped)')
    arg('list_file', metavar='LIST',
        help='Tab-separated list with (at least) two columns <readname> and <haplotype> (can be gzipped). '
        'Currently, the two haplotypes have to be named H1 and H2 (or none). Alternatively, the '
        'output of the "haplotag" command can be used (4 columns), and this is required for using '
        'the "--only-largest-block" option (need phaseset and chromosome info).')
# fmt: on


def validate(args, parser):
    if (args.output_h1 is None) and (args.output_h2 is None) and (args.output_untagged is None):
        parser.error(
            "Nothing to be done since neither --output-h1 nor --output-h2 nor --output-untagged are given."
        )


def select_reads_in_largest_phased_blocks(block_sizes, block_to_readnames):
    """
    :param block_sizes:
    :param block_to_readnames:
    :return:
    """
    selected_reads = set()
    logger.info("Determining largest blocks/phasesets per chromosome")
    for chromosome, block_counts in block_sizes.items():
        block_name, reads_in_block = block_counts.most_common(1)[0]
        logger.info(
            "Chromosome: {} - Phaseset: {} - Tagged reads: {}".format(
                chromosome, block_name, reads_in_block
            )
        )
        selected_reads = selected_reads.union(set(block_to_readnames[(chromosome, block_name)]))
    logger.info(
        "Total number of haplo-tagged reads in all largest phased blocks: {}".format(
            len(selected_reads)
        )
    )
    return selected_reads


def process_haplotag_list_file(
    haplolist, line_parser, haplotype_to_int, only_largest_blocks, discard_unknown_reads
):
    """
    :param haplolist:
    :param line_parser:
    :param haplotype_to_int:
    :param only_largest_blocks:
    :param discard_unknown_reads:
    :return:
    """

    if not haplolist.readline().startswith("#"):
        haplolist.seek(0)

    # needed to determine largest phased block
    block_sizes = defaultdict(Counter)
    # for later removal of reads not in largest phased block;
    # since this can grow quite a bit, only fill if needed
    blocks_to_readnames = defaultdict(set)

    # this set should not be too large given
    # that the haplotag list file contains only
    # a subset of the reads in the input FASTQ/BAM
    known_reads = set()

    readname_to_haplotype = defaultdict(int)

    # No. of reads in the file
    total_reads = 0

    for line in haplolist:
        readname, haplo_name, phaseset, chromosome = line_parser(line)
        total_reads += 1
        try:
            haplo_num = haplotype_to_int[haplo_name]
        except KeyError:
            logger.error(
                "Mapping the haplotype name to the corresponding haplotype "
                "number failed. Currently, the haplotype name in the haplotag "
                "list file has to be one of: none, H1, H2. The value that triggered "
                f"the error was: {haplo_name}"
            )
            raise
        if haplo_num == 0:
            if discard_unknown_reads:
                known_reads.add(readname)

            # TODO the below does not save any memory because
            # the moment a nonexisting key in the defaultdict is accessed
            # it is added, so it’ll just use the momery later.

            # Some "trickery" here:
            # Haplotype 0 means untagged;
            # the return value of a defaultdict(int)
            # is zero for unknown keys, so no need to store
            # anything unless "--discard-unknown-reads" is True,
            # in which case we need to know all read names
            continue
        readname_to_haplotype[readname] = haplo_num
        if only_largest_blocks:
            block_sizes[chromosome][phaseset] += 1
            blocks_to_readnames[(chromosome, phaseset)].add(readname)

    # No. of reads that were tagged with H1 or H2
    tagged_reads = len(readname_to_haplotype)
    untagged_reads = total_reads - tagged_reads
    logger.info("Total number of reads in haplotag list: %d", total_reads)
    logger.info("Total number of haplo-tagged reads: %d", tagged_reads)
    logger.info("Total number of untagged reads: %d", untagged_reads)

    if discard_unknown_reads:
        known_reads.update(readname_to_haplotype)
        num_known_reads = len(known_reads)
        assert total_reads == num_known_reads, (
            "Total number of reads is not equal to number of known reads:"
            f" {total_reads} != {num_known_reads}"
        )

    if only_largest_blocks:
        selected_reads = select_reads_in_largest_phased_blocks(block_sizes, blocks_to_readnames)
        readname_to_haplotype = defaultdict(
            int, {k: readname_to_haplotype[k] for k in selected_reads}
        )
        num_removed_reads = total_reads - len(readname_to_haplotype)
        logger.info(
            "Number of reads removed / reads not overlapping largest phased blocks: %d",
            num_removed_reads,
        )

    return readname_to_haplotype, known_reads


def _two_column_parser(line):
    cols = line.strip().split("\t")[:2]
    return cols[0], cols[1], None, None


def _four_column_parser(line):
    return line.strip().split("\t")[:4]


def _bam_iterator(bam_file):
    """
    :param bam_file:
    :return:
    """
    for record in bam_file:
        qlen = record.query_length
        if qlen > 0:
            yield record.query_name, qlen, record
        else:
            inferred_qlen = record.infer_query_length()
            if inferred_qlen is not None:
                yield record.query_name, inferred_qlen, record
            else:
                yield record.query_name, 0, record


def _fastq_string_iterator(fastq_file):
    """
    Explicit casting to string because pysam does not seem to
    have a writer for FASTQ files - note that this relies
    on opening all compressed files in "text" mode

    :param fastq_file:
    :return:
    """
    for record in fastq_file:
        yield record.name, len(record.sequence), str(record) + "\n"


def check_haplotag_list_information(haplotag_list, exit_stack):
    """
    Check if the haplotag list file has at least 4 columns
    (assumed to be read name, haplotype, phaseset, chromosome),
    or at least 2 columns (as above). Fails if the haplotag file
    is not tab-separated. Return suitable parser for format

    :param haplotag_list: Tab-separated file with at least 2 or 4 columns
    :param exit_stack:
    :return:
    """
    haplo_list = exit_stack.enter_context(xopen(haplotag_list, threads=0))
    first_line = haplo_list.readline().strip()
    # rewind to make sure a header-less file is processed correctly
    haplo_list.seek(0)
    has_chrom_info = False
    try:
        _, _, _, _ = first_line.split("\t")[:4]
        line_parser = _four_column_parser
    except ValueError:
        try:
            _, _ = first_line.split("\t")[:2]
            line_parser = _two_column_parser
        except ValueError:
            raise ValueError(
                "First line of haplotag list file does not have "
                "at least 2 columns, or it is not tab-separated: {}".format(first_line)
            )
    else:
        has_chrom_info = True
    return haplo_list, has_chrom_info, line_parser


def initialize_io_files(reads_file, output_h1, output_h2, output_untagged, exit_stack):
    """
    :param reads_file:
    :param output_h1:
    :param output_h2:
    :param output_untagged:
    :param exit_stack:
    :return:
    """
    potential_fastq_extensions = ["fastq", "fastq.gz", "fastq.gzip" "fq", "fq.gz" "fq.gzip"]
    input_format = detect_file_format(reads_file)
    if input_format is None:
        # TODO: this is a heuristic, need to extend utils::detect_file_format
        if any(reads_file.endswith(ext) for ext in potential_fastq_extensions):
            input_format = "FASTQ"
        else:
            raise ValueError(
                "Undetected file format for input reads. "
                "Expecting BAM or FASTQ (gzipped): {}".format(reads_file)
            )
    elif input_format == "BAM":
        pass
    elif input_format in ["VCF", "CRAM"]:
        raise ValueError(
            "Input file format detected as: {} " "Currently, only BAM and FASTQ is supported."
        )
    else:
        # this means somebody changed utils::detect_file_format w/o
        # checking for usage throughout the code
        raise ValueError(
            "Unexpected file format for input reads: {} - "
            "Expecting BAM or FASTQ (gzipped)".format(input_format)
        )

    output_writers = []
    if input_format == "BAM":
        input_reader = exit_stack.enter_context(
            pysam.AlignmentFile(
                reads_file,
                mode="rb",
                check_sq=False,  # I guess this is needed for unaligned PacBio native files
            )
        )
        input_iter = _bam_iterator

        for outfile in [output_untagged, output_h1, output_h2]:
            output_writers.append(
                exit_stack.enter_context(
                    pysam.AlignmentFile(
                        os.devnull if outfile is None else outfile, mode="wb", template=input_reader
                    )
                )
            )
    elif input_format == "FASTQ":
        # raw or gzipped is both handled by PySam
        input_reader = exit_stack.enter_context(pysam.FastxFile(reads_file))
        output_mode = "wb"
        if not (reads_file.endswith(".gz") or reads_file.endswith(".gzip")):
            output_mode = "w"
        input_iter = _fastq_string_iterator
        for outfile in [output_untagged, output_h1, output_h2]:
            output_writers.append(
                exit_stack.enter_context(
                    open(os.devnull, output_mode) if outfile is None else xopen(outfile, "w")
                )
            )
    else:
        # and this means I overlooked something...
        raise ValueError(f"Unhandled file format for input reads: {input_format}")
    return input_reader, input_iter, output_writers


def write_read_length_histogram(length_counts, path):
    h1 = length_counts[1]
    h2 = length_counts[2]
    untag = length_counts[0]
    all_read_lengths = sorted(itertools.chain(*(h1.keys(), h2.keys(), untag.keys())))
    with xopen(path, "w") as tsv_file:
        print("#length", "count-untagged", "count-h1", "count-h2", sep="\t", file=tsv_file)
        for rlen in all_read_lengths:
            print(rlen, untag[rlen], h1[rlen], h2[rlen], sep="\t", file=tsv_file)


def run_split(
    reads_file,
    list_file,
    output_h1=None,
    output_h2=None,
    output_untagged=None,
    add_untagged=False,
    only_largest_block=False,
    discard_unknown_reads=False,
    read_lengths_histogram=None,
):
    timers = StageTimer()
    timers.start("split-run")

    with ExitStack() as stack:
        timers.start("split-init")

        # TODO: obviously this won't work for more than two haplotypes
        haplotype_to_int = {"none": 0, "H1": 1, "H2": 2}

        haplo_list, has_haplo_chrom_info, line_parser = check_haplotag_list_information(
            list_file, stack
        )

        if only_largest_block:
            logger.debug(
                'User selected "--only-largest-block", this requires chromosome '
                "and phaseset information to be present in the haplotag list file."
            )
            if not has_haplo_chrom_info:
                raise ValueError(
                    "The haplotag list file does not contain phaseset and chromosome "
                    "information, which is required to select only reads from the "
                    "largest phased block. Columns 3 and 4 are missing."
                )

        timers.start("split-process-haplotag-list")

        readname_to_haplotype, known_reads = process_haplotag_list_file(
            haplo_list, line_parser, haplotype_to_int, only_largest_block, discard_unknown_reads
        )
        if discard_unknown_reads:
            logger.debug(
                "User selected to discard unknown reads, i.e., ignore all reads "
                "that are not part of the haplotag list input file."
            )
            assert (
                len(known_reads) > 0
            ), "No known reads in input set - would discard everything, this is probably wrong"
            missing_reads = len(known_reads)
        else:
            missing_reads = -1

        timers.stop("split-process-haplotag-list")

        input_reader, input_iterator, output_writers = initialize_io_files(
            reads_file, output_h1, output_h2, output_untagged, stack
        )

        timers.stop("split-init")

        histogram_data = {
            0: Counter(),
            1: Counter(),
            2: Counter(),
        }

        # holds count statistics about total processed reads etc.
        read_counter = Counter()

        process_haplotype = {
            0: output_untagged is not None or add_untagged,
            1: output_h1 is not None,
            2: output_h2 is not None,
        }

        timers.start("split-iter-input")

        for read_name, read_length, record in input_iterator(input_reader):
            read_counter["total_reads"] += 1
            if discard_unknown_reads and read_name not in known_reads:
                read_counter["unknown_reads"] += 1
                continue
            read_haplotype = readname_to_haplotype[read_name]
            if not process_haplotype[read_haplotype]:
                read_counter["skipped_reads"] += 1
                continue
            histogram_data[read_haplotype][read_length] += 1
            read_counter[read_haplotype] += 1

            output_writers[read_haplotype].write(record)
            if read_haplotype == 0 and add_untagged:
                output_writers[1].write(record)
                output_writers[2].write(record)

            if discard_unknown_reads:
                missing_reads -= 1
                if missing_reads == 0:
                    logger.info("All known reads processed - cancel processing...")
                    break

        timers.stop("split-iter-input")

        if read_lengths_histogram is not None:
            timers.start("split-length-histogram")
            write_read_length_histogram(histogram_data, read_lengths_histogram)
            timers.stop("split-length-histogram")

    timers.stop("split-run")

    logger.info("\n== SUMMARY ==")
    logger.info("Total reads processed: {}".format(read_counter["total_reads"]))
    logger.info(f'Number of output reads "untagged": {read_counter[0]}')
    logger.info(f"Number of output reads haplotype 1: {read_counter[1]}")
    logger.info(f"Number of output reads haplotype 2: {read_counter[2]}")
    logger.info("Number of unknown (dropped) reads: {}".format(read_counter["unknown_reads"]))
    logger.info(
        "Number of skipped reads (per user request): {}".format(read_counter["skipped_reads"])
    )

    logger.info(
        "Time for processing haplotag list: {} sec".format(
            round(timers.elapsed("split-process-haplotag-list"), 3)
        )
    )

    logger.info(
        "Time for total initial setup: {} sec".format(round(timers.elapsed("split-init"), 3))
    )

    logger.info(
        "Time for iterating input reads: {} sec".format(
            round(timers.elapsed("split-iter-input"), 3)
        )
    )

    if read_lengths_histogram is not None:
        logger.info(
            "Time for creating histogram output: {} sec".format(
                round(timers.elapsed("split-length-histogram"), 3)
            )
        )

    logger.info("Total run time: {} sec".format(round(timers.elapsed("split-run"), 3)))


def main(args):
    run_split(**vars(args))



from typing import Optional, Set

from whatshap.core import ReadSet

def readselection(
    readset: ReadSet,
    max_cov: int,
    preferred_source_ids: Optional[Set[int]] = ...,
    bridging: bool = ...,
) -> Set[int]: ...



from collections import namedtuple
from typing import Dict, Iterable, Optional, Tuple, List, Set, Sequence, Iterator

Variant = namedtuple("Variant", "position allele quality")

class NumericSampleIds:
    def __getitem__(self, sample: str) -> int: ...
    def __len__(self) -> int: ...
    def __str__(self) -> str: ...
    def freeze(self) -> None: ...
    def inverse_mapping(self) -> Dict[int, str]: ...

class Read:
    def __init__(
        self,
        name: Optional[str] = ...,
        mapq: int = ...,
        source_id: int = ...,
        sample_id: int = ...,
        reference_start: int = ...,
        BX_tag: Optional[str] = ...,
    ): ...
    @property
    def mapqs(self) -> Tuple[int]: ...
    @property
    def name(self) -> str: ...
    @property
    def source_id(self) -> int: ...
    @property
    def sample_id(self) -> int: ...
    @property
    def reference_start(self) -> int: ...
    @property
    def BX_tag(self) -> str: ...
    def __iter__(self) -> Iterator[Variant]: ...
    def __len__(self) -> int: ...
    def __getitem__(self, key: int) -> Variant: ...
    def __setitem__(self, index: int, variant: Variant): ...
    def __contains__(self, position: int) -> bool: ...
    def add_variant(self, position: int, allele: int, quality: int) -> None: ...
    def add_mapq(self, mapq: int) -> None: ...
    def sort(self) -> None: ...
    def is_sorted(self) -> bool: ...
    def has_BX_tag(self) -> bool: ...

class ReadSet:
    def add(self, read: Read) -> None: ...
    def __iter__(self) -> Iterator[Read]: ...
    def __len__(self) -> int: ...
    def __getitem__(self, key: int) -> Read: ...
    def sort(self) -> None: ...
    def subset(self, reads_to_select: Iterable[int]) -> ReadSet: ...
    def get_positions(self) -> List[int]: ...

class PedigreeDPTable:
    def __init__(
        self,
        readset: ReadSet,
        recombcost: Sequence[int],
        pedigree: Pedigree,
        distrust_genotypes: bool = ...,
        positions: Optional[Iterable[int]] = ...,
    ): ...
    def get_super_reads(self) -> Tuple[List[ReadSet], List[int]]: ...
    def get_optimal_cost(self) -> int: ...
    def get_optimal_partitioning(self) -> List[int]: ...

class Pedigree:
    def __init__(self, numeric_sample_ids: NumericSampleIds): ...
    def add_individual(
        self,
        id: str,
        genotypes: Iterable[Genotype],
        genotype_likelihoods: Optional[Iterable[PhredGenotypeLikelihoods]] = ...,
    ) -> None: ...
    def add_relationship(self, father_id: int, mother_id: int, child_id: int) -> None: ...
    @property
    def variant_count(self) -> int: ...
    def genotype(self, sample_id: int, variant_index: int) -> Genotype: ...
    def genotype_likelihoods(
        self, sample_id: int, variant_index: int
    ) -> Optional[PhredGenotypeLikelihoods]: ...
    def __len__(self) -> int: ...

class PhredGenotypeLikelihoods:
    def __init__(self, gl: Iterable[float], ploidy: int = ..., nr_alleles: int = ...): ...
    def __getitem__(self, genotype: Genotype) -> float: ...
    def __len__(self) -> int: ...
    def __iter__(self) -> Iterator[float]: ...
    def __eq__(self, other) -> bool: ...
    def genotypes(self) -> List[Genotype]: ...

def binomial_coefficient(n: int, k: int) -> int: ...

class Genotype:
    def __init__(self, alleles: List[int]): ...
    def is_none(self) -> bool: ...
    def get_index(self) -> int: ...
    def as_vector(self) -> List[int]: ...
    def is_homozygous(self) -> bool: ...
    def is_diploid_and_biallelic(self) -> bool: ...
    def get_ploidy(self) -> int: ...
    def __eq__(self, other) -> bool: ...
    def __ne__(self, other) -> bool: ...
    def __lt__(self, other) -> bool: ...

def get_max_genotype_ploidy() -> int: ...
def get_max_genotype_alleles() -> int: ...

class GenotypeDPTable:
    def __init__(
        self,
        numeric_sample_ids: NumericSampleIds,
        readset: ReadSet,
        recombcost: int,
        pedigree: Pedigree,
        positions: Optional[Iterable[int]] = ...,
    ): ...
    def get_genotype_likelihoods(self, sample_id: int, pos: int) -> PhredGenotypeLikelihoods: ...

def compute_genotypes(
    readset: ReadSet, positions: Optional[Iterable[int]] = ...
) -> Tuple[List[Genotype], List[Tuple[float, float, float]]]: ...

class HapChatCore:
    def __init__(self, readset: ReadSet): ...
    def get_length(self) -> int: ...
    def get_super_reads(self) -> Tuple[List[ReadSet], List[int]]: ...
    def get_optimal_cost(self) -> int: ...
    def get_optimal_partitioning(self) -> List[int]: ...



import time
from collections import defaultdict
from contextlib import contextmanager


class StageTimer:
    """Measure run times of multiple non-overlapping stages of a program"""

    def __init__(self):
        self._start = dict()
        self._elapsed = defaultdict(float)
        self._overall_start_time = time.time()

    def start(self, stage):
        """Start measuring elapsed time for a stage"""
        self._start[stage] = time.time()

    def stop(self, stage):
        """Stop measuring elapsed time for a stage."""
        t = time.time() - self._start[stage]
        assert t > 0
        self._elapsed[stage] += t
        del self._start[stage]
        return t

    def elapsed(self, stage):
        """
        Return total time spent in a stage, which is the sum of the time spans
        between calls to start() and stop(). If the timer is currently running,
        its current invocation is not counted.
        """
        return self._elapsed[stage]

    def sum(self):
        """Return sum of all times"""
        return sum(self._elapsed.values())

    def total(self):
        return time.time() - self._overall_start_time

    @contextmanager
    def __call__(self, stage):
        self.start(stage)
        yield
        self.stop(stage)

    def iterate(self, stage, iterator):
        """Measure iterator runtime"""

        self.start(stage)
        for item in iterator:
            self.stop(stage)
            yield item
            self.start(stage)
        self.stop(stage)



import itertools
import logging
from collections import defaultdict
from .polyphase_solver import HaploThreader
from math import ceil, log
from scipy.stats import binom

from whatshap.polyphaseutil import get_position_map

logger = logging.getLogger(__name__)


def run_threading(
    readset,
    clustering,
    ploidy,
    max_cluster_gap=10,
    genotypes=None,
    error_rate=0.05,
):
    """
    Main method for the threading stage of the polyploid phasing algorithm. Takes the following input:

    readset -- The fragment matrix to phase
    clustering -- A list of clusters. Each cluster is a list of read ids, indicating which reads it contains. Every read can
                  only be present in one cluster.
    ploidy -- Number of haplotypes to phase
    max_cluster_gap -- Number of allowed consecutive zero-coverage positions for a cluster to still
                       be considered for the threading
    genotypes -- Desired genotype as dictionary (allele -> multiplicity), as list over positions.
                 None, to not force genotypes
    error_rate -- Estimation of allele error rate for the reads. Only relevant when forcing genotypes

    """

    # compute auxiliary data
    index, rev_index = get_position_map(readset)
    num_vars = len(rev_index)
    ad, cons_lists = get_allele_depths(readset, clustering, ploidy)
    cov_map = select_clusters(ad, ploidy, max_cluster_gap)

    # compute threading through the clusters
    affine_switch_cost = ceil(compute_readlength_snp_distance_ratio(readset) / 1.0)
    paths = compute_threading_path(
        readset,
        cov_map,
        ad,
        ploidy,
        switch_cost=4 * affine_switch_cost,
        affine_switch_cost=affine_switch_cost,
        max_cluster_gap=max_cluster_gap,
    )
    assert len(paths) == num_vars

    # determine haplotypes/alleles for each position
    haplotypes = compute_haplotypes(paths, cov_map, cons_lists, ploidy)

    # enforce genotypes
    if genotypes:
        haplotypes = force_genotypes(
            paths, haplotypes, genotypes, clustering, cov_map, ad, error_rate
        )

    return (paths, haplotypes)


def compute_readlength_snp_distance_ratio(readset):
    length = 0
    for read in readset:
        length += len(read)
    return length / len(readset)


def compute_threading_path(
    readset,
    cov_map,
    ad,
    ploidy,
    switch_cost=32.0,
    affine_switch_cost=8.0,
    max_cluster_gap=10,
):
    """
    Runs the threading algorithm for the haplotypes using the given costs for switches. The normal switch cost is the
    cost per haplotype, that switches clusters from one position to another (non matching coverage on single position
    is always cost 1.0). The affine switch cost is an additional offset for every position, where a switch occurs.
    These additional costs encourage the threading algorithm to summarize multiple switches on consecutive positions
    into one big switch.
    """

    logger.debug(f"Computing threading paths with switch cost {switch_cost} ..")

    # run threader
    row_limit = 16 * 2**ploidy if ploidy > 6 else 0
    threader = HaploThreader(ploidy, switch_cost, affine_switch_cost, max_cluster_gap, row_limit)

    path = threader.computePathsBlockwise([0], cov_map, ad)

    return path


def compute_haplotypes(path, cov_map, consensus_lists, ploidy):
    """
    Fills each haplotypes using the computed clusters and their consensus lists
    """
    haplotypes = [[] for _ in range(ploidy)]

    for pos in range(len(path)):
        cnts = defaultdict(int)
        for i in range(ploidy):
            cid = path[pos][i]
            if cid in consensus_lists[pos]:
                allele = consensus_lists[pos][cid][cnts[cid]]
            else:
                allele = -1
            cnts[cid] += 1
            haplotypes[i].append(allele)

    return haplotypes


def force_genotypes(path, haplotypes, genotypes, clustering, cov_map, allele_depths, error_rate):
    num_vars = len(path)
    for pos in range(num_vars):
        # count allele occurences
        alleles = {a for a in genotypes[pos]}
        present = defaultdict(int)
        for h in haplotypes:
            present[h[pos]] += 1
            alleles.add(h[pos])

        # detect abundances and shortages
        abundant_alleles, lacking_alleles = dict(), dict()
        alleles_to_insert, affected_positions = [], []
        for a in alleles:
            if a not in genotypes[pos]:
                genotypes[pos][a] = 0
            diff = present[a] - genotypes[pos][a]
            if diff > 0:
                # slots with abundant allele could change, in total we want the lower count
                abundant_alleles[a] = diff
                alleles_to_insert += [a for _ in range(genotypes[pos][a])]
                for p in range(len(path[pos])):
                    if haplotypes[p][pos] == a:
                        affected_positions.append(p)
            elif diff < 0:
                # slots with short alleles can stay, we need to insert the missing amount
                lacking_alleles[a] = -diff
                alleles_to_insert += [a for _ in range(-diff)]

        affected_positions.sort()
        alleles_to_insert.sort()

        if len(abundant_alleles) == 0:
            continue

        # for all permutations, pick the one best fitting to allele depths of clusters
        clusts = cov_map[pos]
        given_config = [haplotypes[h][pos] for h in range(len(haplotypes))]
        best_config = given_config
        best_likelihood = -float("inf")
        for perm in set(list(itertools.permutations(alleles_to_insert))):
            # build next config (given + affected slots permuted)
            newconfig = given_config[:]
            for i in range(len(perm)):
                newconfig[affected_positions[i]] = perm[i]

            # compute likelihood how well newconfig explains the observed allele depths of the clusters
            log_likelihood = 0

            # compute allele depth fraction for each cluster
            for clust in clusts:
                allele_mult = {a: 0 for a in alleles}
                clust_mult = 0
                for slot in range(len(path[pos])):
                    if path[pos][slot] == clust:
                        allele_mult[newconfig[slot]] += 1
                        clust_mult += 1
                if clust_mult > 0:
                    # compute likelihood that observed allele depth was produced by assigned alleles per cluster
                    total_depth = sum(allele_depths[pos][clust].values())
                    for a in alleles:
                        allele_mult[a] /= clust_mult
                        allele_mult[a] = (
                            allele_mult[a] * (1 - error_rate) + (1 - allele_mult[a]) * error_rate
                        )
                        observed_depth = (
                            0
                            if a not in allele_depths[pos][clust]
                            else allele_depths[pos][clust][a]
                        )
                        log_likelihood += log(
                            binom.pmf(observed_depth, total_depth, allele_mult[a])
                        )

            if log_likelihood > best_likelihood:
                best_likelihood = log_likelihood
                best_config = newconfig

        for h in range(len(haplotypes)):
            haplotypes[h][pos] = best_config[h]

    return haplotypes


def select_clusters(allele_depths, ploidy, max_gap):
    """
    For every position, computes a list of relevant clusters for the threading
    algorithm. Relevant means, that the relative coverage is at least 1/8 of
    what a single haplotype is expected to have for the given ploidy. Apart
    from that, at least <ploidy> and at most <ploidy + 2> many clusters are
    selected to avoid exponential blow-up.
    """
    cov_map = [[] for _ in range(len(allele_depths))]
    for pos in range(len(allele_depths)):
        sorted_cids = sorted(
            ((cid, sum(allele_depths[pos][cid].values())) for cid in allele_depths[pos]),
            key=lambda x: x[1],
            reverse=True,
        )
        total_cov = sum(e[1] for e in sorted_cids)
        cut_off = min(len(sorted_cids), ploidy + 2)
        cov_map[pos].append(sorted_cids[0][0])
        for (cid, cov) in sorted_cids[1:cut_off]:
            if cov / total_cov < (1.0 / (8.0 * ploidy)) and cov_map[pos]:
                break
            else:
                cov_map[pos].append(cid)

    # re-add clusters missing on at most max_gap intermediate positions
    cut_off = ploidy + 2
    for pos in range(1, len(cov_map) - 1):
        for cid in cov_map[pos - 1]:
            if len(cov_map[pos]) >= cut_off:
                break
            if cid in cov_map[pos]:
                continue
            if any(
                [cid in cov_map[pos + k + 1] for k in range(min(max_gap, len(cov_map) - pos - 1))]
            ):
                cov_map[pos].append(cid)
                allele_depths[pos][cid] = dict()

    for sub in cov_map:
        sub.sort()

    return cov_map


def get_allele_depths(readset, clustering, ploidy):
    """
    Returns a list, which for every position contains a list (representing the clusters) of dictionaries containing the allele depths.
    Additionally computes a consensus list per position per cluster, such that the first k elements represent the alleles of this
    cluster, if it was selected with multiplicity k.
    ad[pos][c_id][al] = number of reads in cluster c_id having allele al at position pos
    Indices are local, i.e. the i-th entry of ad is the entry for the i-th position that occurs in the readset.
    """
    # Map genome positions to [0,l)
    index = {}
    rev_index = []
    num_vars = 0
    for position in readset.get_positions():
        index[position] = num_vars
        rev_index.append(position)
        num_vars += 1

    # stores allele depth per position and cluster
    ad = [dict() for pos in range(num_vars)]
    cons_lists = [dict() for pos in range(num_vars)]

    # count alleles
    for c_id, cluster in enumerate(clustering):
        for read in cluster:
            for var in readset[read]:
                pos = index[var.position]
                if c_id not in ad[pos]:
                    ad[pos][c_id] = dict()
                if var.allele not in ad[pos][c_id]:
                    ad[pos][c_id][var.allele] = 0
                ad[pos][c_id][var.allele] += 1

    # compute allele lists
    for pos in range(num_vars):
        for c_id in ad[pos]:
            cons_lists[pos][c_id] = []
            cnts = defaultdict(int)
            for i in range(ploidy):
                max_cnt = 0
                max_al = 0
                for al in ad[pos][c_id]:
                    cnt = ad[pos][c_id][al] / (1 + cnts[al])
                    if cnt > max_cnt:
                        max_cnt = cnt
                        max_al = al
                cons_lists[pos][c_id].append(max_al)
                cnts[max_al] += 1

    return ad, cons_lists



from whatshap.cli.unphase import run_unphase


def test_unphase(tmpdir):
    out = tmpdir.join("out.vcf")
    run_unphase("tests/data/phased-via-mixed-HP-PS.vcf", str(out))
    with open("tests/data/unphased.vcf") as f:
        expected = f.read()
    assert expected == out.read_text(encoding="ascii")


def test_unphase_string_typed_ps(tmpdir):
    # Ensure a VCF with PS tags of type String (although against VCF spec) can be read
    run_unphase("tests/data/string_typed_ps_tag.vcf", str(tmpdir.join("out.vcf")))



"""
Print phasing statistics of a single VCF file
"""
import logging
from collections import defaultdict
from contextlib import ExitStack
import dataclasses
from statistics import median
from typing import List, Tuple, Optional

from ..vcf import VcfReader

logger = logging.getLogger(__name__)


# fmt: off
def add_arguments(parser):
    add = parser.add_argument
    add("--gtf", default=None, help="Write phased blocks to GTF file.")
    add("--sample", metavar="SAMPLE", help="Name of the sample "
        "to process. If not given, use first sample found in VCF.")
    add("--chr-lengths", metavar="FILE",
        help="Override chromosome lengths in VCF with those from FILE (one line per chromosome, "
        "tab separated '<chr> <length>'). Lengths are used to compute NG50 values.")
    add("--tsv", metavar="FILE", help="Write statistics in tab-separated value format to FILE")
    add("--only-snvs", default=False, action="store_true", help="Only process SNVs "
        "and ignore all other variants.")
    add("--block-list", metavar="FILE", help="Write list of all blocks to FILE (one block per line)")
    add("--chromosome", dest="chromosomes", metavar="CHROMOSOME", default=[], action="append",
        help="Name of chromosome to process. If not given, all chromosomes in the "
        "input VCF are considered. Can be used multiple times")
    add("vcf", metavar="VCF", help="Phased VCF file")
# fmt: on


def validate(args, parser):
    pass


class PhasedBlock:
    def __init__(self, chromosome=None):
        self.phases = {}
        self.leftmost_variant = None
        self.rightmost_variant = None
        self.chromosome = chromosome

    def add(self, variant, phase):
        if len(self.phases) == 0:
            self.leftmost_variant = variant
            self.rightmost_variant = variant
        else:
            if variant < self.leftmost_variant:
                self.leftmost_variant = variant
            if self.rightmost_variant < variant:
                self.rightmost_variant = variant
        self.phases[variant] = phase

    def span(self):
        """Returns the length of the covered genomic region in bp."""
        return self.rightmost_variant.position - self.leftmost_variant.position

    def variants(self):
        return list(sorted(self.phases.keys()))

    def count_snvs(self):
        return sum(int(variant.is_snv()) for variant in self.phases)

    def split(self, split_left: int, split_right: int) -> Tuple["PhasedBlock", "PhasedBlock"]:
        """Split this phaseblock in two, based on given positions. The first phaseblock will contain
        the variants to the left of split_left and the second the variants to the right of split_right."""
        assert split_left <= split_right
        left_block = PhasedBlock(chromosome=self.chromosome)
        right_block = PhasedBlock(chromosome=self.chromosome)
        for variant, phase in self.phases.items():
            if variant.position < split_left:
                left_block.add(variant, phase)
            elif variant.position > split_right:
                right_block.add(variant, phase)
        return left_block, right_block

    def __repr__(self):
        return f"PhasedBlock({str(self.phases)})"

    def __len__(self):
        return len(self.phases)

    def __lt__(self, other):
        return (self.leftmost_variant, self.rightmost_variant) < (
            other.leftmost_variant,
            other.rightmost_variant,
        )


class GtfWriter:
    def __init__(self, file):
        self._file = file

    def write(self, chromosome, start, stop, name):
        """
        Write a feature to the GTF. start is 0-based.
        """
        assert start < stop
        print(
            chromosome,
            "Phasing",
            "exon",
            start + 1,
            stop,
            ".",
            "+",
            ".",
            f'gene_id "{name}"; transcript_id "{name}.1";',
            sep="\t",
            file=self._file,
        )


@dataclasses.dataclass
class DetailedStats:
    variants: int
    phased: int
    unphased: int
    singletons: int
    blocks: int
    variant_per_block_median: float
    variant_per_block_avg: float
    variant_per_block_min: int
    variant_per_block_max: int
    variant_per_block_sum: int
    bp_per_block_median: float
    bp_per_block_avg: float
    bp_per_block_min: int
    bp_per_block_max: int
    bp_per_block_sum: int
    heterozygous_variants: int
    heterozygous_snvs: int
    phased_snvs: int
    block_n50: float


def n50(lengths: List[int], target_length: Optional[int] = None) -> int:
    if target_length is None:
        target_length = sum(lengths)

    lengths.sort(reverse=True)
    total = 0
    for length in lengths:
        total += length
        if total >= 0.5 * target_length:
            return length
    return 0


def compute_ng50(blocks, chr_lengths):
    chromosomes = {b.chromosome for b in blocks}
    target_length = 0
    for chromosome in sorted(chromosomes):
        try:
            target_length += chr_lengths[chromosome]
        except KeyError:
            logger.warning(
                "Not able to compute NG50 because length of contig '%s' not available", chromosome
            )
            return float("nan")

    block_lengths = [b.rightmost_variant.position - b.leftmost_variant.position for b in blocks]
    return n50(block_lengths, target_length=target_length)


class PhasingStats:
    def __init__(self):
        self.blocks = []
        self.unphased = 0
        self.variants = 0
        self.heterozygous_variants = 0
        self.heterozygous_snvs = 0
        self.phased_snvs = 0

    def __iadd__(self, other):
        self.blocks.extend(other.blocks)
        self.unphased += other.unphased
        self.variants += other.variants
        self.heterozygous_variants += other.heterozygous_variants
        self.heterozygous_snvs += other.heterozygous_snvs
        self.phased_snvs += other.phased_snvs
        return self

    def add_blocks(self, blocks):
        self.blocks.extend(blocks)

    def add_unphased(self, unphased: int = 1):
        self.unphased += unphased

    def add_variants(self, variants: int):
        self.variants += variants

    def add_heterozygous_variants(self, variants: int):
        self.heterozygous_variants += variants

    def add_heterozygous_snvs(self, snvs: int):
        self.heterozygous_snvs += snvs

    def get_nonoverlapping_blocks(self) -> List[PhasedBlock]:
        """Split phase blocks into nonoverlapping subblocks"""
        pos_sorted_blocks = sorted(
            self.blocks, key=lambda b: (b.chromosome, b.leftmost_variant.position), reverse=True
        )

        # filter out blocks with only one variant
        pos_sorted_blocks = [b for b in pos_sorted_blocks if len(b) > 1]

        # iterate over blocks and split if overlapping until no blocks remain.
        split_blocks = []
        while pos_sorted_blocks:
            block = pos_sorted_blocks.pop()
            if pos_sorted_blocks:
                block_end = block.rightmost_variant.position
                next_block = pos_sorted_blocks[-1]
                next_block_start = next_block.leftmost_variant.position
                next_block_end = next_block.rightmost_variant.position

                # Check if next block overlapps current. If so split the current block.
                if (block_end > next_block_start) and (block.chromosome == next_block.chromosome):
                    block, new_block = block.split(next_block_start, next_block_end)

                    # Update sorting if right-side block is added.
                    if len(new_block) > 1:
                        pos_sorted_blocks.append(new_block)
                        pos_sorted_blocks = sorted(
                            pos_sorted_blocks,
                            key=lambda b: (b.chromosome, b.leftmost_variant.position),
                            reverse=True,
                        )

                    # Skip the left-side block if is it too short after splitting
                    if len(block) < 2:
                        continue
            split_blocks.append(block)

        return split_blocks

    def get(self, chr_lengths=None):
        """Return DetailedStats"""
        block_sizes = sorted(len(block) for block in self.blocks)
        n_singletons = sum(1 for size in block_sizes if size == 1)
        block_sizes = [size for size in block_sizes if size > 1]
        block_lengths = sorted(block.span() for block in self.blocks if len(block) > 1)
        phased_snvs = sum(block.count_snvs() for block in self.blocks if len(block) > 1)
        if block_sizes:
            # Split interleaved blocks to avoid inflating NG50
            split_blocks = self.get_nonoverlapping_blocks()
            return DetailedStats(
                variants=self.variants,
                phased=sum(block_sizes),
                unphased=self.unphased,
                singletons=n_singletons,
                blocks=len(block_sizes),
                variant_per_block_median=median(block_sizes),
                variant_per_block_avg=sum(block_sizes) / len(block_sizes),
                variant_per_block_min=block_sizes[0],
                variant_per_block_max=block_sizes[-1],
                variant_per_block_sum=sum(block_sizes),
                bp_per_block_median=median(block_lengths),
                bp_per_block_avg=sum(block_lengths) / len(block_lengths),
                bp_per_block_min=block_lengths[0],
                bp_per_block_max=block_lengths[-1],
                bp_per_block_sum=sum(block_lengths),
                heterozygous_variants=self.heterozygous_variants,
                heterozygous_snvs=self.heterozygous_snvs,
                phased_snvs=phased_snvs,
                block_n50=compute_ng50(split_blocks, chr_lengths)
                if chr_lengths is not None
                else float("nan"),
            )
        else:
            return DetailedStats(
                variants=self.variants,
                phased=0,
                unphased=self.unphased,
                singletons=n_singletons,
                blocks=0,
                variant_per_block_median=float("nan"),
                variant_per_block_avg=float("nan"),
                variant_per_block_min=0,
                variant_per_block_max=0,
                variant_per_block_sum=0,
                bp_per_block_median=float("nan"),
                bp_per_block_avg=float("nan"),
                bp_per_block_min=0,
                bp_per_block_max=0,
                bp_per_block_sum=0,
                heterozygous_variants=self.heterozygous_variants,
                heterozygous_snvs=self.heterozygous_snvs,
                phased_snvs=0,
                block_n50=float("nan"),
            )

    def print(self, chr_lengths=None):
        stats = self.get(chr_lengths)
        WIDTH = 21
        print("Variants in VCF:".rjust(WIDTH), f"{stats.variants:8d}")
        print(
            "Heterozygous:".rjust(WIDTH),
            f"{stats.heterozygous_variants:8d} ({stats.heterozygous_snvs:8d} SNVs)",
        )
        print("Phased:".rjust(WIDTH), f"{stats.phased:8d} ({stats.phased_snvs:8d} SNVs)")
        print("Unphased:".rjust(WIDTH), f"{stats.unphased:8d}", "(not considered below)")
        print("Singletons:".rjust(WIDTH), f"{stats.singletons:8d}", "(not considered below)")
        print("Blocks:".rjust(WIDTH), f"{stats.blocks:8d}")
        print()
        print("Block sizes (no. of variants)")
        print(
            "Median block size:".rjust(WIDTH),
            f"{stats.variant_per_block_median:11.2f} variants",
        )
        print(
            "Average block size:".rjust(WIDTH),
            f"{stats.variant_per_block_avg:11.2f} variants",
        )
        print("Largest block:".rjust(WIDTH), f"{stats.variant_per_block_max:8d}    variants")
        print("Smallest block:".rjust(WIDTH), f"{stats.variant_per_block_min:8d}    variants")
        print()
        print("Block lengths (basepairs)")
        print("Sum of lengths:".rjust(WIDTH), f"{stats.bp_per_block_sum:8d}    bp")
        print("Median block length:".rjust(WIDTH), f"{stats.bp_per_block_median:11.2f} bp")
        print("Average block length:".rjust(WIDTH), f"{stats.bp_per_block_avg:11.2f} bp")
        print("Longest block:".rjust(WIDTH), f"{stats.bp_per_block_max:8d}    bp")
        print("Shortest block:".rjust(WIDTH), f"{stats.bp_per_block_min:8d}    bp")
        print("Block NG50:".rjust(WIDTH), f"{stats.block_n50:8.0f}    bp")
        assert stats.phased + stats.unphased + stats.singletons == stats.heterozygous_variants


def parse_chr_lengths(filename):
    chr_lengths = {}
    with open(filename) as f:
        for line in f:
            fields = line.split("\t")
            assert len(fields) == 2
            chr_lengths[fields[0]] = int(fields[1])
    return chr_lengths


def parse_variant_tables(vcf_reader, chromosomes=None):
    """
    Parse variant_tables from vcf_reader. If chromosomes are given and VCF is indexed,
    theses are accessed by direct lookup.
    """
    if chromosomes and vcf_reader.index_exists():
        for chromosome in chromosomes:
            yield vcf_reader.fetch(chromosome)
    else:
        yield from vcf_reader


def run_stats(
    vcf,
    sample=None,
    gtf=None,
    tsv=None,
    block_list=None,
    only_snvs=False,
    chromosomes=None,
    chr_lengths=None,
):
    gtfwriter = tsv_file = block_list_file = None
    with ExitStack() as stack:
        if gtf:
            gtf_file = stack.enter_context(open(gtf, "wt"))
            gtfwriter = GtfWriter(gtf_file)
        if tsv:
            tsv_file = stack.enter_context(open(tsv, "w"))
        if block_list:
            block_list_file = stack.enter_context(open(block_list, "w"))

        vcf_reader = VcfReader(vcf, phases=True, indels=not only_snvs)
        if len(vcf_reader.samples) == 0:
            logger.error("Input VCF does not contain any sample")
            return 1
        else:
            logger.info(f"Found {len(vcf_reader.samples)} sample(s) in input VCF")
        if sample:
            if sample in vcf_reader.samples:
                sample = sample
            else:
                logger.error(f"Requested sample ({sample}) not found")
                return 1
        else:
            sample = vcf_reader.samples[0]
            logger.info(f"Reporting results for sample {sample}")

        if chr_lengths:
            chr_lengths = parse_chr_lengths(chr_lengths)
            logger.info("Read length of %d chromosomes from %s", len(chr_lengths), chr_lengths)
        else:
            chr_lengths = {
                chrom: contig.length
                for chrom, contig in vcf_reader.contigs.items()
                if contig.length is not None
            }
            if not chr_lengths:
                logger.warning(
                    "VCF header does not contain contig lengths, cannot compute NG50. "
                    "Consider using --chr-lengths"
                )

        if tsv_file:
            field_names = [f.name for f in dataclasses.fields(DetailedStats)]
            print("#sample", "chromosome", "file_name", *field_names, sep="\t", file=tsv_file)

        if block_list_file:
            print(
                "#sample",
                "chromosome",
                "phase_set",
                "from",
                "to",
                "variants",
                sep="\t",
                file=block_list_file,
            )

        print(f"Phasing statistics for sample {sample} from file {vcf}")
        total_stats = PhasingStats()
        chromosome_count = 0
        given_chromosomes = chromosomes
        seen_chromosomes = set()
        for variant_table in parse_variant_tables(vcf_reader, given_chromosomes):
            if given_chromosomes:
                seen_chromosomes.add(variant_table.chromosome)
                if variant_table.chromosome not in given_chromosomes:
                    continue
            chromosome_count += 1
            chromosome = variant_table.chromosome
            stats = PhasingStats()
            print(f"---------------- Chromosome {chromosome} ----------------")
            genotypes = variant_table.genotypes_of(sample)
            phases = variant_table.phases_of(sample)
            assert len(genotypes) == len(phases) == len(variant_table.variants)
            blocks = defaultdict(PhasedBlock)
            prev_block_id = None
            prev_block_fragment_start = None
            prev_block_fragment_end = None
            for variant, genotype, phase in zip(variant_table.variants, genotypes, phases):
                stats.add_variants(1)
                if genotype.is_homozygous():
                    continue
                stats.add_heterozygous_variants(1)
                if variant.is_snv():
                    stats.add_heterozygous_snvs(1)
                if phase is None:
                    stats.add_unphased()
                else:
                    # a phased variant
                    blocks[phase.block_id].add(variant, phase)
                    if gtfwriter:
                        if prev_block_id is None:
                            prev_block_fragment_start = variant.position
                            prev_block_fragment_end = variant.position + 1
                            prev_block_id = phase.block_id
                        else:
                            if prev_block_id != phase.block_id:
                                gtfwriter.write(
                                    chromosome,
                                    prev_block_fragment_start,
                                    prev_block_fragment_end,
                                    prev_block_id,
                                )
                                prev_block_fragment_start = variant.position
                                prev_block_id = phase.block_id
                            prev_block_fragment_end = variant.position + 1

            # Add chromosome information to each block. This is needed to
            # sort blocks later when we compute NG50s
            for block_id, block in blocks.items():
                block.chromosome = chromosome

            if gtfwriter and prev_block_id is not None:
                gtfwriter.write(
                    chromosome, prev_block_fragment_start, prev_block_fragment_end, prev_block_id
                )

            if block_list_file:
                block_ids = sorted(blocks.keys())
                for block_id in block_ids:
                    print(
                        sample,
                        chromosome,
                        block_id,
                        blocks[block_id].leftmost_variant.position + 1,
                        blocks[block_id].rightmost_variant.position + 1,
                        len(blocks[block_id]),
                        sep="\t",
                        file=block_list_file,
                    )

            stats.add_blocks(blocks.values())
            stats.print(chr_lengths)
            if tsv_file:
                print(sample, chromosome, vcf, sep="\t", end="\t", file=tsv_file)
                print(*dataclasses.astuple(stats.get(chr_lengths)), sep="\t", file=tsv_file)

            total_stats += stats

            if given_chromosomes and set(given_chromosomes) <= seen_chromosomes:
                break

        if chromosome_count > 1:
            print("---------------- ALL chromosomes (aggregated) ----------------")
            total_stats.print(chr_lengths)
            if tsv_file:
                print(sample, "ALL", vcf, sep="\t", end="\t", file=tsv_file)
                print(*dataclasses.astuple(total_stats.get(chr_lengths)), sep="\t", file=tsv_file)


def main(args):
    run_stats(**vars(args))



import sys
import resource
import logging

from whatshap.bam import (
    AlignmentFileNotIndexedError,
    EmptyAlignmentFileError,
    SampleNotFoundError,
    ReferenceNotFoundError,
)
from whatshap.variants import ReadSetReader, ReadSetError
from whatshap.utils import IndexedFasta, FastaNotIndexedError, detect_file_format
from whatshap.core import ReadSet
from whatshap.vcf import VcfReader

logger = logging.getLogger(__name__)


class CommandLineError(Exception):
    """An anticipated command-line error occurred. This ends up as a user-visible error message"""


def open_readset_reader(*args, **kwargs):
    try:
        readset_reader = ReadSetReader(*args, **kwargs)
    except OSError as e:
        raise CommandLineError(e)
    except AlignmentFileNotIndexedError as e:
        raise CommandLineError(
            "The file '{}' is not indexed. Please create the appropriate BAM/CRAM "
            'index with "samtools index"'.format(e.args[0])
        )
    except EmptyAlignmentFileError as e:
        raise CommandLineError(
            "No reads could be retrieved from '{}'. If this is a CRAM file, possibly the "
            "reference could not be found. Try to use --reference=... or check your "
            "$REF_PATH/$REF_CACHE settings".format(e.args[0])
        )
    return readset_reader


class PhasedInputReader:
    def __init__(
        self,
        bam_or_vcf_paths,
        reference,
        numeric_sample_ids,
        ignore_read_groups,
        indels,
        **kwargs,  # passed to ReadSetReader constructor
    ):
        self._bam_paths, self._vcf_paths = self._split_input_file_list(bam_or_vcf_paths)

        # TODO exit stack!
        self._numeric_sample_ids = numeric_sample_ids
        self._fasta = self._open_reference(reference) if reference else None

        vcf_readers = [VcfReader(f, indels=indels, phases=True) for f in self._vcf_paths]

        self._vcf_readers = vcf_readers
        self._ignore_read_groups = ignore_read_groups

        self._readset_reader = open_readset_reader(
            self._bam_paths, reference, numeric_sample_ids, **kwargs
        )
        if not self._vcf_readers:
            self._vcfs = []
        else:
            self._vcfs = None  # None means uninitialized, call .read_vcf() first

    def __enter__(self):
        return self

    def __exit__(self, *args):
        if self._fasta is not None:
            self._fasta.close()

    @property
    def has_vcfs(self):
        return bool(self._vcf_paths)

    @property
    def has_alignments(self) -> bool:
        """Whether any of the input files are BAM or CRAM"""
        return bool(self._bam_paths)

    @staticmethod
    def _split_input_file_list(paths):
        bams = []
        vcfs = []
        for path in paths:
            try:
                file_format = detect_file_format(path)
            except OSError as e:
                raise CommandLineError(e)
            if file_format in ("BAM", "CRAM"):
                bams.append(path)
            elif file_format == "VCF":
                vcfs.append(path)
            else:
                raise CommandLineError(f"Unable to determine type of input file {path!r}")
        return bams, vcfs

    @staticmethod
    def _open_reference(path):
        try:
            indexed_fasta = IndexedFasta(path)
        except OSError as e:
            raise CommandLineError(f"Error while opening FASTA reference file: {e}")
        except FastaNotIndexedError as e:
            raise CommandLineError(
                f"An index file (.fai) for the reference FASTA '{e.args[0]}' "
                "could not be found. Please create one with "
                "'samtools faidx'."
            )
        return indexed_fasta

    def read_vcfs(self):
        # Read phase information provided as VCF files, if provided.
        # TODO: do this chromosome- and/or sample-wise on demand to save memory.
        self._vcfs = []
        for reader in self._vcf_readers:
            # create dict mapping chromosome names to VariantTables
            m = dict()
            logger.info("Reading phased blocks from %r", reader.path)
            for variant_table in reader:
                m[variant_table.chromosome] = variant_table
            self._vcfs.append(m)

    def read(self, chromosome, variants, sample, *, read_vcf=True, regions=None):
        """
        Return a pair (readset, vcf_source_ids) where readset is a sorted ReadSet.

        Set read_vcf to False to not read phased blocks from the VCFs
        """
        readset_reader = self._readset_reader
        for_sample = f"for sample {sample!r} " if not self._ignore_read_groups else ""
        logger.debug(
            "Reading alignments %son chromosome %s and detecting alleles ...",
            for_sample,
            chromosome,
        )
        try:
            reference = self._fasta[chromosome] if self._fasta else None
        except KeyError:
            raise CommandLineError(
                f"Chromosome {chromosome!r} present in VCF file, "
                f"but not in the reference FASTA {self._fasta.filename!r}"
            )
        bam_sample = None if self._ignore_read_groups else sample
        try:
            readset = readset_reader.read(chromosome, variants, bam_sample, reference, regions)
        except SampleNotFoundError:
            logger.warning("Sample %r not found in any BAM/CRAM file.", bam_sample)
            readset = ReadSet()
        except ReadSetError as e:
            raise CommandLineError(e)
        except ReferenceNotFoundError:
            if chromosome.startswith("chr"):
                alternative = chromosome[3:]
            else:
                alternative = "chr" + chromosome
            message = f"The chromosome {chromosome!r} was not found in the BAM/CRAM file."
            if readset_reader.has_reference(alternative):
                message += f" Found {alternative!r} instead"
            raise CommandLineError(message)

        vcf_source_ids = set()
        if read_vcf:
            # TODO this is a bit clumsy
            if self._vcfs is None:
                raise ValueError("call PhasedInputReader.read_vcfs() first")
            # Add phasing information from VCF files, if present
            sample_id = self._numeric_sample_ids[sample]
            for i, vcf in enumerate(self._vcfs):
                if chromosome in vcf:
                    variant_table = vcf[chromosome]
                    source_id = readset_reader.n_paths + i
                    vcf_source_ids.add(source_id)
                    for read in variant_table.phased_blocks_as_reads(
                        sample, variants, source_id, sample_id
                    ):
                        readset.add(read)

        # TODO is this necessary?
        for read in readset:
            read.sort()
        readset.sort()

        logger.info(
            "Found %d reads covering %d variants", len(readset), len(readset.get_positions())
        )
        return readset, vcf_source_ids


def log_memory_usage(include_children=False):
    if sys.platform == "linux":
        if include_children:
            memory_kb = (
                resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
                + resource.getrusage(resource.RUSAGE_CHILDREN).ru_maxrss
            )
        else:
            memory_kb = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
        logger.info("Maximum memory usage: %.3f GB", memory_kb / 1e6)



"""
Generate candidate SNP positions.
"""

import pysam
import sys
import re
import pyfaidx
from collections import defaultdict
import datetime
import logging

logger = logging.getLogger(__name__)


# fmt: off
def add_arguments(parser):
    add = parser.add_argument
    add('ref', metavar='REF', help='FASTA with reference genome')
    add('bam', metavar='BAM', help='BAM file')
    add('--minabs', metavar='MIN_ABS', default=3, type=int,
        help='Minimum absolute ALT depth to call a SNP (default: %(default)s).')
    add('--minrel', metavar='MIN_REL', default=0.25, type=float,
        help='Minimum relative ALT depth to call a SNP (default: %(default)s).')
    add('--multi-allelics', default=False, action='store_true',
        help='Also output multi-allelic sites, if not given only the best ALT allele is reported (if unique).')
    add('--sample', metavar='SAMPLE', default='sample',
        help='Put this sample column into VCF (default: output sites-only VCF).')
    add('--chromosome', dest='chromosome', metavar='CHROMOSOME', default=None,
        help='Name of chromosome to process. If not given, all chromosomes are processed.')
    add('-o', '--output', default=sys.stdout, help='Output VCF file.')
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        '--pacbio', dest='datatype', action='store_const', const='pacbio',
        help='Input is PacBio. Sets minrel=0.25 and minabs=3.')
    group.add_argument(
        '--nanopore', dest='datatype', action='store_const', const='nanopore',
        help='Input is Nanopore. Sets minrel=0.4 and minabs=3.')
    group.add_argument(
        '--illumina', dest='datatype', action='store_const', const='illumina',
        help='Input is Illumina. Sets minrel=0.25 and minabs=3.')
# fmt: on


def validate(args, parser):
    pass


def run_find_snv_candidates(
    ref,
    bam,
    minabs=3,
    minrel=0.25,
    multi_allelics=False,
    datatype=None,
    sample="sample",
    chromosome=None,
    output=sys.stdout,
):
    outfile = output
    if output != sys.stdout:
        outfile = open(output, "w")
    if datatype == "pacbio":
        minabs = 3
        minrel = 0.25
    if datatype == "nanopore":
        minabs = 3
        minrel = 0.4
    if datatype == "illumina":
        minabs = 3
        minrel = 0.25
    print(minabs, minrel)
    fasta = pyfaidx.Fasta(ref, as_raw=True)
    print("##fileformat=VCFv4.2", file=outfile)
    print("##fileDate={}".format(datetime.datetime.now().strftime("%Y%m%d")), file=outfile)
    print('##FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype">', file=outfile)
    print('##FILTER=<ID=PASS,Description="All filters passed">', file=outfile)
    header_columns = ["#CHROM", "POS", "ID", "REF", "ALT", "QUAL", "FILTER", "INFO"]
    if sample is not None:
        header_columns += ["FORMAT", sample]
    print(*header_columns, sep="\t", file=outfile)

    re_nucleotide = re.compile("[ACGTNacgtn]")
    re_indel = re.compile("[-\\+]([0-9]+)")
    re_ref = re.compile("[,\\.]")
    re_ignore = re.compile("([\\$\\*]|\\^.)")

    bamfile = pysam.AlignmentFile(bam, "rb")
    for pileupcolumn in bamfile.pileup(
        contig=chromosome, min_mapping_quality=20, min_base_quality=5
    ):
        try:
            pileup = "".join(
                pileupcolumn.get_query_sequences(mark_matches=True, mark_ends=True, add_indels=True)
            )
        except AssertionError:
            # TODO: seems to happen when there are large insertions/deletions
            continue
        position = pileupcolumn.reference_pos + 1
        chromosome = pileupcolumn.reference_name
        i = 0
        bases = defaultdict(int)
        n = 0
        ref = fasta[chromosome][position - 1].upper()
        if ref == "N":
            continue
        while i < len(pileup):
            m = re_nucleotide.match(pileup[i:])
            if m is not None:
                bases[pileup[i].upper()] += 1
                n += 1
                i += 1
                continue
            m = re_indel.match(pileup[i:])
            if m is not None:
                l = int(m.group(1))
                skip = (m.end() - m.start()) + l
                i += skip
                continue
            m = re_ref.match(pileup[i:])
            if m is not None:
                bases[ref] += 1
                n += 1
                i += 1
                continue
            m = re_ignore.match(pileup[i:])
            if m is not None:
                skip = m.end() - m.start()
                i += skip
                continue
            assert False
        ref_count = bases[ref]
        alts = []
        for base, count in bases.items():
            if base == ref:
                continue
            if (count >= minabs) and (count / (count + ref_count) >= minrel):
                alts.append((count, base))
        alts.sort(reverse=True)
        if len(alts) > 0:
            columns = [chromosome, position, ".", ref, ".", ".", "PASS", "."]
            if sample is not None:
                columns += ["GT", "."]
            if multi_allelics:
                columns[4] = ",".join(base for count, base in alts)
            else:
                # Do we have two equally supported ALT alleles
                if len(alts) > 1 and (alts[0][0] == alts[1][0]):
                    columns[4] = "N"
                    continue
                else:
                    columns[4] = alts[0][1]
            print(*columns, sep="\t", file=outfile)
    if output != sys.stdout:
        outfile.close()


def main(args):
    run_find_snv_candidates(**vars(args))



import ast
import sys
import pkgutil
import importlib
import logging

import whatshap.cli as cli_package
from . import __version__
from .args import HelpfulArgumentParser
from .cli import CommandLineError


logger = logging.getLogger(__name__)


class NiceFormatter(logging.Formatter):
    """
    Do not prefix "INFO:" to info-level log messages (but do it for all other
    levels).

    Based on http://stackoverflow.com/a/9218261/715090 .
    """

    def format(self, record):
        if record.levelno != logging.INFO:
            record.msg = f"{record.levelname}: {record.msg}"
        return super().format(record)


def setup_logging(debug):
    """
    Set up logging. If debug is True, then DEBUG level messages are printed.
    """
    handler = logging.StreamHandler()
    handler.setFormatter(NiceFormatter())
    root = logging.getLogger()
    root.addHandler(handler)
    root.setLevel(logging.DEBUG if debug else logging.INFO)


def main(argv=None):
    if argv is None:
        argv = sys.argv[1:]
    subcommand_name = get_subcommand_name(argv)
    module = importlib.import_module("." + subcommand_name, cli_package.__name__)

    parser = HelpfulArgumentParser(description=__doc__, prog="whatshap")
    parser.add_argument("--version", action="version", version="%(prog)s " + __version__)
    parser.add_argument("--debug", action="store_true", default=False, help="Print debug messages")
    subparsers = parser.add_subparsers()
    subparser = subparsers.add_parser(
        subcommand_name,
        help=module.__doc__.strip().split("\n", maxsplit=1)[0],
        description=module.__doc__,
    )
    module.add_arguments(subparser)
    args = parser.parse_args(argv)
    setup_logging(args.debug)

    if hasattr(module, "validate"):
        module.validate(args, subparser)
    del args.debug
    try:
        module.main(args)
    except CommandLineError as e:
        logger.error("whatshap error: %s", str(e))
        logger.debug("Command line error. Traceback:", exc_info=True)
        sys.exit(1)


def get_subcommand_name(arguments) -> str:
    """
    Parse arguments to find out which subcommand was requested.

    This sets up a minimal ArgumentParser with the correct help strings.

    Because help is obtained from a module’s docstring, but importing each module
    makes startup slow, the modules are only parsed with the ast module and
    not fully imported at this stage.

    Return:
        subcommand name
    """
    parser = HelpfulArgumentParser(description=__doc__, prog="whatshap")
    parser.add_argument("--version", action="version", version=__version__)
    subparsers = parser.add_subparsers()

    for module_name, docstring in cli_modules(cli_package):
        help = docstring.strip().split("\n", maxsplit=1)[0].replace("%", "%%")
        subparser = subparsers.add_parser(
            module_name, help=help, description=docstring, add_help=False
        )
        subparser.set_defaults(module_name=module_name)
    args, _ = parser.parse_known_args(arguments)
    module_name = getattr(args, "module_name", None)
    if module_name is None:
        parser.error("Please provide the name of a subcommand to run")
    return module_name


def cli_modules(package):
    """
    Yield (module_name, docstring) tuples for all modules in the given package.
    """
    modules = pkgutil.iter_modules(package.__path__)
    for module in modules:
        spec = importlib.util.find_spec(package.__name__ + "." + module.name)
        with open(spec.origin) as f:
            mod_ast = ast.parse(f.read())
        docstring = ast.get_docstring(mod_ast, clean=False)
        yield module.name, docstring


if __name__ == "__main__":
    main()



import math
from whatshap.core import (
    ReadSet,
    Pedigree,
    NumericSampleIds,
    PhredGenotypeLikelihoods,
    GenotypeDPTable,
)
from whatshap.testhelpers import (
    string_to_readset,
    canonic_index_to_biallelic_gt,
    canonic_index_list_to_biallelic_gt_list,
)


def compare_to_expected(dp_forward_backward, positions, expected=None, genotypes=None):
    # check if computed likelihoods are equal to expected ones (if given)
    if expected is not None:
        for i in range(len(positions)):
            likelihoods = dp_forward_backward.get_genotype_likelihoods("individual0", i)
            print(likelihoods, expected[i], i)
            assert likelihoods == expected[i]

    # check if likeliest genotype is equal to expected genotype
    for i in range(len(positions)):
        likelihoods = dp_forward_backward.get_genotype_likelihoods("individual0", i)
        # find likeliest genotype
        max_val = -1
        max_geno = -1
        for genotype in likelihoods.genotypes():
            # likelihood should not be nan
            assert not math.isnan(likelihoods[genotype])
            if likelihoods[genotype] > max_val:
                max_val = likelihoods[genotype]
                max_geno = genotype

        print(
            "genotype likelihoods for position", i, likelihoods, " likeliest genotype: ", max_geno
        )

        if genotypes is not None:
            assert max_geno == genotypes[i]


def test_genotyping_empty_readset():
    rs = ReadSet()
    genotypes = canonic_index_list_to_biallelic_gt_list([1, 1])
    recombcost = [1, 1]
    numeric_sample_ids = NumericSampleIds()
    pedigree = Pedigree(numeric_sample_ids)
    genotype_likelihoods = [None, None]
    pedigree.add_individual("individual0", genotypes, genotype_likelihoods)
    _ = GenotypeDPTable(numeric_sample_ids, rs, recombcost, pedigree)


def check_genotyping_single_individual(
    reads, weights=None, expected=None, genotypes=None, scaling=None, genotype_priors=None
):
    # 0) set up read set
    readset = string_to_readset(s=reads, w=weights, scale_quality=scaling)
    positions = readset.get_positions()

    # 1) Genotype using forward backward algorithm
    recombcost = [1] * len(positions)
    numeric_sample_ids = NumericSampleIds()
    pedigree = Pedigree(numeric_sample_ids)
    genotype_likelihoods = [PhredGenotypeLikelihoods([1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0])] * len(
        positions
    )

    if genotype_priors is not None:
        genotype_likelihoods = genotype_priors

    pedigree.add_individual(
        "individual0",
        [canonic_index_to_biallelic_gt(1) for i in range(len(positions))],
        genotype_likelihoods,
    )
    dp_forward_backward = GenotypeDPTable(numeric_sample_ids, readset, recombcost, pedigree)

    # check the results
    compare_to_expected(dp_forward_backward, positions, expected, genotypes)


# TODO: when using non-uniform transitions, the pedigree results are not equal (?)
# 2) Phase using PedMEC code for trios with two "empty" individuals (i.e. having no reads)
#    recombcost = [1] * len(positions) # recombination costs 1, should not occur
#    numeric_sample_ids = NumericSampleIds()
#    pedigree = Pedigree(numeric_sample_ids)
#    pedigree.add_individual('individual0', [1] * len(positions),genotype_likelihoods)
#    pedigree.add_individual('individual1', [1] * len(positions), [PhredGenotypeLikelihoods(1/3.0,1/3.0,1/3.0)] * len(positions))
#    pedigree.add_individual('individual2', [1] * len(positions), [PhredGenotypeLikelihoods(1/3.0,1/3.0,1/3.0)] * len(positions))
#    pedigree.add_relationship('individual0', 'individual1', 'individual2')
#    dp_forward_backward = GenotypeDPTable(numeric_sample_ids,readset,recombcost,pedigree)

# TODO
# check the results
#    compare_to_expected(dp_forward_backward, positions, expected, genotypes)

# first 5 tests compare to genotype likelihoods computed manually, with (non-)uniform priors


def test_geno_exact1():
    reads = """
          11
           01
        """

    expected_likelihoods = [
        PhredGenotypeLikelihoods([0.06666666666666667, 0.3333333333333333, 0.6]),
        PhredGenotypeLikelihoods([0.20930232558139536, 0.5813953488372093, 0.20930232558139536]),
        PhredGenotypeLikelihoods([0.06666666666666667, 0.3333333333333333, 0.6]),
    ]
    genotypes = canonic_index_list_to_biallelic_gt_list([2, 1, 2])
    check_genotyping_single_individual(reads, None, expected_likelihoods, genotypes, 10)


def test_geno_exact2():
    reads = """
        11
        11
        """
    weights = """
        11
        11
        """

    expected_likelihoods = [
        PhredGenotypeLikelihoods([0.00914139256727894, 0.25040580948312685, 0.7404527979495942]),
        PhredGenotypeLikelihoods([0.00914139256727894, 0.25040580948312685, 0.7404527979495942]),
    ]
    genotypes = canonic_index_list_to_biallelic_gt_list([2, 2])
    check_genotyping_single_individual(reads, weights, expected_likelihoods, genotypes, 10)


def test_geno_exact3():
    reads = """
          01
          11
        """

    expected_likelihoods = [
        PhredGenotypeLikelihoods([0.22163406214039125, 0.5567318757192175, 0.22163406214039125]),
        PhredGenotypeLikelihoods([0.009896432681242807, 0.18849252013808976, 0.8016110471806674]),
    ]
    check_genotyping_single_individual(reads, None, expected_likelihoods, None, 10)


def test_geno_priors1():
    reads = """
          01
          11
        """

    prior_likelihoods = [
        PhredGenotypeLikelihoods([0.1, 0.8, 0.1]),
        PhredGenotypeLikelihoods([0.1, 0.2, 0.7]),
    ]
    expected_likelihoods = [
        PhredGenotypeLikelihoods([0.04257892641700095, 0.9148421471659981, 0.04257892641700095]),
        PhredGenotypeLikelihoods([0.0016688611936185199, 0.05208684202468078, 0.9462442967817007]),
    ]
    check_genotyping_single_individual(
        reads, None, expected_likelihoods, None, 10, prior_likelihoods
    )


def test_geno_priors2():
    reads = """
            11
             01
             """

    prior_likelihoods = [
        PhredGenotypeLikelihoods([0, 0.5, 0.5]),
        PhredGenotypeLikelihoods([0.25, 0.5, 0.25]),
        PhredGenotypeLikelihoods([0.1, 0.4, 0.5]),
    ]
    expected_likelihoods = [
        PhredGenotypeLikelihoods([0.0, 0.35714285714285715, 0.6428571428571429]),
        PhredGenotypeLikelihoods([0.1323529411764706, 0.7352941176470589, 0.1323529411764706]),
        PhredGenotypeLikelihoods([0.015151515151515152, 0.30303030303030304, 0.6818181818181818]),
    ]
    check_genotyping_single_individual(
        reads, None, expected_likelihoods, None, 10, prior_likelihoods
    )


# check if exprected genotype predictions are made
def test_geno1():
    reads = """
    1111111111
    0000011111
    """

    genotypes = canonic_index_list_to_biallelic_gt_list([1, 1, 1, 1, 1, 2, 2, 2, 2, 2])
    check_genotyping_single_individual(reads, None, None, genotypes, 10)


def test_geno2():
    reads = """
    101
    101
    101
    101
    100
    100
    100
    100
    """

    genotypes = canonic_index_list_to_biallelic_gt_list([2, 0, 1])
    check_genotyping_single_individual(reads, None, None, genotypes, 10)


def test_geno3():
    reads = """
    111011
    110110
    110 10
    110110
    101110
    000 00
    01000
    000010
    100100
    """

    genotypes = canonic_index_list_to_biallelic_gt_list([1, 1, 0, 1, 1, 0])
    check_genotyping_single_individual(reads, None, None, genotypes, 10)


def test_geno4():
    reads = """
      1  11010
      00 00101
      001 01110
       1    111
    """

    check_genotyping_single_individual(reads, None, None, None, 10)


def test_geno5():
    reads = """
      0             0
      110111111111
      00100
           0001000000
           000
            10100
                  101
    """
    genotypes = canonic_index_list_to_biallelic_gt_list(
        [1, 1, 1, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1, 0, 1]
    )
    check_genotyping_single_individual(reads, None, None, genotypes, 10)


def test_geno6():
    reads = """
        0100000000000
        0100010000000
        1110000000010
        0100000000000
        0101000001000
        0100010   000
        0 10000000100
        1111111011100
        0100111010011
        1111111000111
        1111110011111
        11110000  000
        1110011011111
        1111001011111
        0111111110  1
        """
    genotypes = canonic_index_list_to_biallelic_gt_list([1, 2, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1])
    check_genotyping_single_individual(reads, None, None, genotypes, 60)


def test_geno7():
    reads = """
        111
        101
        111
        101
        010
        000
        010
        000
        """
    genotypes = canonic_index_list_to_biallelic_gt_list([1, 1, 1])
    check_genotyping_single_individual(reads, None, None, genotypes, 60)


def test_geno8():
    reads = """
    11
    11
    10
    """
    genotypes = canonic_index_list_to_biallelic_gt_list([2, 1])
    check_genotyping_single_individual(reads, None, None, genotypes, 10)


def test_geno9():
    reads = """
    111
    111
    010
    010
       001
       001
       101
       101
    """
    genotypes = canonic_index_list_to_biallelic_gt_list([1, 2, 1, 1, 0, 2])
    check_genotyping_single_individual(reads, None, None, genotypes, 10)


def test_geno_10():
    reads = """
    001100
    000000
    000000
    110011
    110011
    111111
         """
    genotypes = canonic_index_list_to_biallelic_gt_list([1, 1, 0, 0, 1, 1])
    genotype_priors = [
        PhredGenotypeLikelihoods([0.1, 0.8, 0.1]),
        PhredGenotypeLikelihoods([0.1, 0.8, 0.1]),
        PhredGenotypeLikelihoods([0.7, 0.2, 0.1]),
        PhredGenotypeLikelihoods([0.7, 0.2, 0.1]),
        PhredGenotypeLikelihoods([0.1, 0.8, 0.1]),
        PhredGenotypeLikelihoods([0.1, 0.8, 0.1]),
    ]
    check_genotyping_single_individual(reads, None, None, genotypes, 50, genotype_priors)


def test_weighted_genotyping1():
    reads = """
      1  11010
      00 00101
      001 01110
       1    111
    """
    weights = """
      2  13112
      11 23359
      223 56789
       2    111
    """

    genotypes = canonic_index_list_to_biallelic_gt_list([1, 1, 2, 1, 1, 1, 2, 1, 1])
    check_genotyping_single_individual(reads, weights, None, genotypes, 10)


def test_weighted_genotyping2():
    reads = """
      111
      101
      111
      101
      010
      000
      010
      000
    """
    weights = """
      999
      999
      999
      999
      999
      999
      999
      999
    """

    # for the second position, each genotype should be equally likely
    expected_likelihoods = [
        PhredGenotypeLikelihoods([0, 1, 0]),
        PhredGenotypeLikelihoods([1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0]),
        PhredGenotypeLikelihoods([0, 1, 0]),
    ]
    check_genotyping_single_individual(reads, weights, expected_likelihoods, None, 100)


def test_weighted_genotyping3():
    reads = """
        0 1
         10
         """
    weights = """
        999
        999
    """
    expected_likelihoods = [
        PhredGenotypeLikelihoods([2.0 / 3.0, 1.0 / 3.0, 0]),
        PhredGenotypeLikelihoods([0, 1.0 / 3.0, 2.0 / 3.0]),
        PhredGenotypeLikelihoods([0, 1, 0]),
    ]
    check_genotyping_single_individual(reads, weights, expected_likelihoods, None, 500)


def test_weighted_genotyping4():
    reads = """
    00  00
    0000
    000
    111
    111101
    111111
    111110
      000
      1111
    """

    weights = """
    11  11
    1111
    111
    111
    111111
    111111
    111111
      111
      1111
    """
    genotypes = canonic_index_list_to_biallelic_gt_list([1, 1, 1, 1, 1, 1])
    check_genotyping_single_individual(reads, weights, None, genotypes, 10)


def test_weighted_genotyping5():
    reads = """
    1111
    1111
    1111
    1111
    1111
    1111
    1111
    1111
    1111
    1111
    1111
    0 00
    00
    0 00
    """

    weights = """
    1111
    1111
    1111
    1111
    1111
    1111
    1111
    1111
    1111
    1111
    1111
    1 11
    1111
    1 11
    """
    genotypes = canonic_index_list_to_biallelic_gt_list([1, 1, 1, 1])
    check_genotyping_single_individual(reads, weights, None, genotypes, 10)


def test_weighted_genotyping6():
    reads = """
        10
        10
         """
    weights = """
        99
        99
    """
    genotype_priors = [
        PhredGenotypeLikelihoods([0.5, 0.5, 0]),
        PhredGenotypeLikelihoods([0, 0.5, 0.5]),
    ]
    expected_likelihoods = [
        PhredGenotypeLikelihoods([0, 1, 0]),
        PhredGenotypeLikelihoods([0, 1, 0]),
    ]
    check_genotyping_single_individual(
        reads, weights, expected_likelihoods, None, 1000, genotype_priors
    )


def test_small_example():
    reads = """
    11111111
    00000000
    """
    genotypes = canonic_index_list_to_biallelic_gt_list([1, 1, 1, 1, 1, 1, 1, 1])
    check_genotyping_single_individual(reads, None, None, genotypes, 1000)
